# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import TextIOBase
from typing import Literal, overload

from rich import print as _richprint

from gurk.lib.utils import typecheck


@typecheck
def newline(*, file: TextIOBase | None = None) -> None:
    """
    Print a newline to the console output.

    :param file: The output file (stdout/stderr). If None, defaults to stdout.
    :type file: TextIOBase | None
    """
    print("", file=file)


@overload
def richprint(
    message: str,
    color: str | None = ...,
    as_str: Literal[False] = ...,
    file: TextIOBase | None = ...,
) -> None: ...


@overload
def richprint(
    message: str,
    color: str | None = ...,
    as_str: Literal[True] = ...,
    file: TextIOBase | None = ...,
) -> str: ...


@typecheck
def richprint(
    message: str,
    color: str | None = None,
    as_str: bool = False,
    file: TextIOBase | None = None,
) -> str | None:
    """
    Print a rich-formatted message with optional color.

    :param message: The message to print
    :type message: str
    :param color: Optional color for the message. If None, defaults to no additional color wrapping.
    :type color: str | None
    :param as_str: If True, return the formatted message as a string instead of printing it
    :type as_str: bool
    :param file: The output file (stdout/stderr). If None, defaults to stdout.
    :type file: TextIOBase | None
    :return: The formatted message if as_str is True, otherwise None
    :rtype: str | None
    """
    if color is not None:
        msg = f"[{color}]{message}[/{color}]"
    else:
        msg = message

    if as_str:
        return msg
    else:
        _richprint(msg, file=file)


@overload
def padded_print(
    text: str,
    color: str = ...,
    total_length: int = ...,
    top: bool = ...,
    bottom: bool = ...,
    as_str: Literal[False] = ...,
    file: TextIOBase | None = ...,
) -> None: ...


@overload
def padded_print(
    text: str,
    color: str = ...,
    total_length: int = ...,
    top: bool = ...,
    bottom: bool = ...,
    as_str: Literal[True] = ...,
    file: TextIOBase | None = ...,
) -> str: ...


@typecheck
def padded_print(
    text: str,
    color: str = "white",
    total_length: int = 128,
    top: bool = True,
    bottom: bool = True,
    as_str: bool = False,
    file: TextIOBase | None = None,
) -> str | None:
    """
    Print text padded with "=" signs to center it within a specified total length.

    :param text: Text to be printed
    :type text: str
    :param color: Color of the text
    :type color: str
    :param total_length: Total length of the printed line including padding
    :type total_length: int
    :param top: Whether to print the top padding line
    :type top: bool
    :param bottom: Whether to print the bottom padding line
    :type bottom: bool
    :param as_str: If True, return the formatted message as a string instead of printing it
    :type as_str: bool
    :param file: The output file (stdout/stderr). If None, defaults to stdout.
    :type file: TextIOBase | None
    :return: The formatted message if as_str is True, otherwise None
    :rtype: str | None
    """
    msg = ""

    # Top bar
    if top:
        msg += (
            richprint("=" * total_length, color=color, file=file, as_str=True)
            + "\n"
        )

    # Calculate how many "=" signs are needed in the middle
    #   Subtract 2 for extra spaces
    remaining_length = total_length - len(text) - 2
    if remaining_length < 0:
        msg += richprint(f"{text}", color=color, file=file, as_str=True)
    else:
        left_pad = remaining_length // 2
        right_pad = remaining_length - left_pad
        msg += richprint(
            f"{'=' * left_pad} {text} {'=' * right_pad}",
            color=color,
            file=file,
            as_str=True,
        )
    # Bottom bar
    if bottom:
        msg += "\n" + richprint(
            "=" * total_length, color=color, file=file, as_str=True
        )

    if as_str:
        return msg
    else:
        _richprint(msg, file=file)



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from functools import cache, wraps
from importlib.metadata import Distribution
from types import GenericAlias
from typing import Any, Literal, TypeGuard, TypeVar, overload

from pydantic import ConfigDict, TypeAdapter, ValidationError, validate_call

from .constants import NO_ANSWERS, PACKAGE_NAME, YES_ANSWERS


@cache
def _is_typecheck_active() -> bool:
    # (Priority) Check environment variable override
    gurk_typecheck = os.getenv("GURK_TYPECHECK")
    if gurk_typecheck in YES_ANSWERS:
        return True
    elif gurk_typecheck in NO_ANSWERS:
        return False
    else:
        # See if the package is installed in editable mode
        direct_url = Distribution.from_name(PACKAGE_NAME).read_text(
            "direct_url.json"
        )
        return (
            json.loads(direct_url).get("dir_info", {}).get("editable", False)
        )


class InputValidationError(Exception):
    """Custom error type for input validation errors in typecheck."""

    pass


def typecheck(
    _func=None,
    /,
    *,
    strict: bool = True,
    extra: Literal["forbid", "allow", "ignore"] = "forbid",
):
    """
    Decorator (or decorator factory) that wraps a function with runtime type checking.

    Can be used as:
      @typecheck                               uses defaults
      @typecheck(strict=False, extra="allow")  custom settings

    :param strict: Whether to enforce strict type checking (e.g., disallowing extra fields in TypedDicts)
    :type strict: bool
    :param extra: How to handle extra fields in TypedDicts ("forbid", "allow", "ignore")
    :type extra: Literal["forbid", "allow", "ignore"]
    :raises InputValidationError: If the input validation fails, with details on the offending inputs
    """

    def decorator(func):
        if not _is_typecheck_active():
            return func

        checked_func = validate_call(
            config=ConfigDict(
                strict=strict, extra=extra, arbitrary_types_allowed=True
            )
        )(func)

        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return checked_func(*args, **kwargs)
            except ValidationError as e:
                messages = []
                for err in e.errors():
                    loc = ".".join(str(x) for x in err["loc"])
                    bad_input = err.get("input", "<missing>")
                    bad_type = type(bad_input).__name__
                    bad_repr = repr(bad_input)
                    if len(bad_repr) > 200:
                        bad_repr = bad_repr[:197] + "..."
                    messages.append(
                        f"{loc}: {err['msg']} (got {bad_type}: {bad_repr})"
                    )
                raise InputValidationError(
                    "Wrong input(s) to '"
                    + func.__name__
                    + "':\n"
                    + "\n".join(messages)
                ) from None

        return wrapper

    if _func is not None:
        return decorator(_func)
    return decorator


T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


@overload
def full_isinstance(
    value: Any,
    expected_type: type[T],
    /,
    *,
    strict: bool = ...,
    extra: str = ...,
) -> TypeGuard[T]: ...


@overload
def full_isinstance(
    value: Any,
    expected_type: type[set[T]],
    /,
    *,
    strict: bool = ...,
    extra: str = ...,
) -> TypeGuard[set[T]]: ...


@overload
def full_isinstance(
    value: Any,
    expected_type: type[list[T]],
    /,
    *,
    strict: bool = ...,
    extra: str = ...,
) -> TypeGuard[list[T]]: ...


@overload
def full_isinstance(
    value: Any,
    expected_type: type[dict[K, V]],
    /,
    *,
    strict: bool = ...,
    extra: str = ...,
) -> TypeGuard[dict[K, V]]: ...


@overload
def full_isinstance(
    value: Any,
    expected_type: type[tuple[T, ...]],
    /,
    *,
    strict: bool = ...,
    extra: str = ...,
) -> TypeGuard[tuple[T, ...]]: ...


@typecheck
def full_isinstance(
    value: Any,
    expected_type: type | GenericAlias,
    /,
    *,
    strict: bool = True,
    extra: str = "forbid",
) -> bool:
    """
    Check if a value is an instance of the expected type.

    :param value: The value to check.
    :type value: Any
    :param expected_type: The expected type (can be a plain type, Union, TypedDict, etc.).
    :type expected_type: type | GenericAlias
    :param strict: Whether to enforce strict type checking (e.g., disallowing extra fields in TypedDicts)
    :type strict: bool
    :param extra: How to handle extra fields in TypedDicts ("forbid", "allow", "ignore")
    :type extra: str
    :return: True if the value matches the expected type, False otherwise.
    :rtype: bool
    """
    adapter = TypeAdapter(expected_type)
    try:
        adapter.validate_python(value, strict=strict, extra=extra)
    except ValidationError:
        return False
    else:
        return True



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shutil
from pathlib import Path
from subprocess import DEVNULL, CalledProcessError, check_call
from venv import EnvBuilder

from gurk.lib.context import get_logger
from gurk.lib.utils import PACKAGE_SRC_PATH, PACKAGE_VENVS_PATH, typecheck


@typecheck
def get_venv_dir(plugin_name: str) -> Path:
    """
    Get the path to the virtual environment directory for a plugin.

    :param plugin_name: Name of the plugin
    :type plugin_name: str
    :return: Path to the virtual environment directory
    :rtype: Path
    """
    return PACKAGE_VENVS_PATH / plugin_name


@typecheck
def venv_exists(venv_name: str) -> bool:
    """
    Check if a virtual environment exists for a plugin.

    :param venv_name: Name of the virtual environment (same as the plugin name)
    :type venv_name: str
    :return: True if the virtual environment exists, False otherwise
    :rtype: bool
    """
    return get_venv_dir(venv_name).is_dir()


@typecheck
def create_venv(venv_name: str, dependencies: list[str]) -> bool:
    """
    Create a virtual environment for a plugin and install its dependencies.

    :param venv_name: Name of the virtual environment (same as the plugin name)
    :type venv_name: str
    :param dependencies: List of dependencies to install
    :type dependencies: list[str]
    :return: True if the virtual environment was created successfully, False otherwise
    :rtype: bool
    """
    # Get logger
    logger = get_logger()

    # Check if venv already exists
    venv_dir = get_venv_dir(venv_name)
    if venv_exists(venv_name):
        logger.error(
            f"Virtual environment for plugin '{venv_name}' already exists at {venv_dir}"
        )
        return False

    # Create venv
    dependencies_str = (
        ("\n- " + "\n- ".join(dependencies)) if dependencies else " None"
    )
    logger.info(
        f"Creating virtual environment for plugin '{venv_name}' "
        f"in {venv_dir} with dependencies:{dependencies_str}"
    )
    EnvBuilder(with_pip=True).create(venv_dir)

    # Install dependencies
    pip_bin = (venv_dir / "bin" / "pip").as_posix()
    all_dependencies = dependencies + [
        "-e",
        PACKAGE_SRC_PATH.parents[1].as_posix(),
    ]
    try:
        check_call([pip_bin, "install", "--upgrade", "pip"], stdout=DEVNULL)
        check_call([pip_bin, "install", *all_dependencies], stdout=DEVNULL)
    except (KeyboardInterrupt, CalledProcessError) as e:
        logger.error(
            f"Failed to install dependencies for virtual environment '{venv_name}': {e}"
        )
        remove_venv(venv_name)
        return False

    logger.success(
        f"Successfully created virtual environment for plugin '{venv_name}'"
    )
    return True


@typecheck
def remove_venv(venv_name: str) -> bool:
    """
    Remove the virtual environment for a plugin, if it exists.

    :param venv_name: Name of the virtual environment (same as the plugin name)
    :type venv_name: str
    :return: True if the virtual environment was removed successfully, False otherwise
    :rtype: bool
    """
    venv_path = get_venv_dir(venv_name)
    if venv_path.is_dir():
        try:
            shutil.rmtree(venv_path)
        except Exception:
            return False
        else:
            return True
    return False


@typecheck
def get_venv_package_version(venv_name: str, package_name: str) -> str | None:
    """
    Get the version of a specific package associated with a virtual environment.

    :param venv_name: Name of the virtual environment (same as the plugin name)
    :type venv_name: str
    :param package_name: Name of the package to get the version for
    :type package_name: str
    :return: The package version if available, None otherwise
    :rtype: str | None
    """
    python_bin = (get_venv_dir(venv_name) / "bin" / "python").as_posix()
    try:
        return check_call(
            [
                python_bin,
                "-c",
                f"from importlib.metadata import version; print(version('{package_name}'))",
            ],
            capture_output=True,
            text=True,
        ).stdout.strip()
    except Exception:
        return None



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import ExitStack

from .logger import DummyLogger, Logger, get_logger
from .registry import RegistryManager


class GurkContext:
    """
    Context manager for Gurk that handles setup and teardown of resources like the plugin registry.
    """

    def __init__(self, *, logger: Logger | None, writable: bool):
        self._logger = logger or DummyLogger()
        self._registry_manager = RegistryManager(writable=writable)
        self._stack = ExitStack()

    def __enter__(self):
        self._stack.enter_context(self._logger)
        self._stack.enter_context(self._registry_manager)
        return self

    def __exit__(self, exc_type, exc, tb):
        return self._stack.__exit__(exc_type, exc, tb)

    @property
    def logger(self) -> Logger:
        return get_logger()



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types
from typing import (
    NotRequired,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
    is_typeddict,
)

T = TypeVar("T")


def fill_typed_dict(dct: dict, tp: type[T]) -> T:
    """
    Recursively fill a TypedDict (or nested structures containing TypedDicts) with default values.
        :NOTE: This function recursively calls itself with different input types, so it is not strictly type-safe.

    :param dct: The dict to fill.
    :type dct: dict
    :param tp: The TypedDict type to use for filling.
    :type tp: type[T]
    :return: The filled dict.
    :rtype: T
    """
    origin = get_origin(tp)
    args = get_args(tp)

    # unwrap NotRequired / Union
    if origin is NotRequired:
        return fill_typed_dict(dct, args[0])
    if origin in (Union, types.UnionType):
        return fill_typed_dict(dct, args[0])

    # TypedDict schema
    if is_typeddict(tp) and isinstance(dct, dict):
        for key, sub_tp in get_type_hints(tp, include_extras=True).items():
            if key not in dct:
                dct[key] = fill_typed_dict(None, sub_tp)
            else:
                dct[key] = fill_typed_dict(dct[key], sub_tp)
        return dct

    # dict[K, V]
    if origin is dict and isinstance(dct, dict):
        val_tp = args[1]
        for k in dct:
            dct[k] = fill_typed_dict(dct[k], val_tp)
        return dct

    # default construction
    if dct is None:
        if origin in (list, set, tuple, dict):
            return origin()
        if is_typeddict(tp):
            return fill_typed_dict({}, tp)
        return tp() if callable(tp) else None

    return dct



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getpass
import os
import subprocess
import sys
from pathlib import Path
from tempfile import NamedTemporaryFile

from rich.markup import escape

from gurk.lib.context import get_logger
from gurk.lib.core.plugins import GurkArgumentParser
from gurk.lib.core.processor import Processor
from gurk.lib.core.scheduler import Scheduler
from gurk.lib.shared.system_info import get_system_info
from gurk.lib.shared.tasks import CustomTaskDictCollection
from gurk.lib.utils import SETUP_DONE_FILE, typecheck


def check_askpass() -> bool:
    """
    Check that 'SUDO_ASKPASS' environment variable is set to a valid executable script.

    :return: True if 'SUDO_ASKPASS' is properly set, False otherwise
    :rtype: bool
    """
    # Get logger
    logger = get_logger()

    # Check 'SUDO_ASKPASS'
    askpass = os.getenv("SUDO_ASKPASS")
    if not askpass:
        logger.debug("'SUDO_ASKPASS' environment variable is not set")
        return False
    elif not Path(askpass).is_file():
        logger.warning(f"'SUDO_ASKPASS' script '{askpass}' not found")
        return False
    elif not os.access(askpass, os.X_OK):
        logger.warning(f"'SUDO_ASKPASS' script '{askpass}' is not executable")
        return False

    return True


def get_sudo_askpass() -> Path:
    """
    Create a temporary sudo askpass script that provides the user's sudo password.

    :return: Path to the temporary askpass script
    :rtype: Path
    """
    # Get logger
    logger = get_logger()

    # Reset sudo permissions
    subprocess.run(["sudo", "-k"])

    # Create temporary askpass file
    with NamedTemporaryFile(mode="w", delete=False) as askpass_file:
        attempts = 3
        while attempts > 0:
            response = logger.ask(
                f"{escape('[gurk]')} password for {getpass.getuser()}", True
            )
            test_response = subprocess.run(
                ["sudo", "-S", "-v"],
                input=response + "\n",
                text=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            if test_response.returncode == 0:
                break
            else:
                if attempts != 1:
                    print("Sorry, try again.", file=sys.stderr)
                attempts -= 1
        else:
            print("gurk: 3 incorrect password attempts", file=sys.stderr)
            raise SystemExit(1)

        askpass_file.write("#!/bin/sh\n" f"echo '{response}'\n")
        askpass_path = askpass_file.name

    os.chmod(askpass_path, 0o700)
    return askpass_path


def prompt_setup() -> None:
    """
    Prompt the user to run setup if it has never been run before.
    """
    # Get logger
    logger = get_logger()

    if not SETUP_DONE_FILE.is_file():
        print(
            "It seems that this is the first time you are running gurk. "
            "It is recommended to run the setup first to ensure all "
            "possible manual steps are taken care of."
        )
        if logger.prompt_bool(
            "Would you like to run the setup now?",
        ):
            from gurk.cli.setup import main as setup_main

            setup_main([], "", "")

        # Mark setup as done
        SETUP_DONE_FILE.parent.mkdir(parents=True, exist_ok=True)
        SETUP_DONE_FILE.touch()


@typecheck
def main(
    option: CustomTaskDictCollection,
    cli_args: list[str],
    parser_base: GurkArgumentParser,
    askpass: str | None = None,
    _captured: list[str] | None = None,
):
    """
    Main entry point for the 'run' command.

    :param option: Run option of the requested gurk plugin
    :type option: CustomTaskDictCollection
    :param cli_args: Command-line arguments to be passed to the resp. tasks
    :type cli_args: list[str]
    :param parser_base: Base argument parser, on which each task parser is built
    :type parser_base: GurkArgumentParser
    :param askpass: (Internal) Sudo askpass path for testing purposes
    :type askpass: str | None
    :param _captured: (Internal) Captured output for testing purposes
    :type _captured: list[str] | None
    :raises Exception: Propagates any exception raised during processing
    """
    # Get logger
    logger = get_logger()

    # Prompt to run the 'setup' command upon first usage
    if logger.can_prompt:
        prompt_setup()

    # Check system information
    system_info = get_system_info()  # Raises exception if incompatible system
    logger.debug(f"System information: {system_info}")

    # Load option and process tasks
    processor = Processor(option, cli_args, parser_base)

    # Check if a prompt is needed to get sudo access
    if not (askpass or check_askpass() or "pytest" in sys.modules):
        if not logger.can_prompt:
            logger.fatal(
                "sudo access is required to run tasks. Please set the "
                "'SUDO_ASKPASS' environment variable or run in interactive mode."
            )
        askpass_prompted = True
    else:
        askpass_prompted = False

    try:
        # Get sudo password
        askpass_path = None  # Set default value in case of early exception
        if askpass_prompted:
            # Prompt for sudo password
            askpass_path = get_sudo_askpass()
        else:
            # Get existing askpass path
            askpass_path = askpass or os.getenv("SUDO_ASKPASS")

        # Schedule and run tasks (where possible, in parallel)
        scheduler = Scheduler(processor.tasks, askpass_path)
        scheduler.run()

        # Save failed tasks (pytest usage)
        if _captured is not None:
            _captured.extend(scheduler.get_results())

    except Exception as e:
        raise e

    finally:
        # Remove temporary sudo askpass file, if it was created via gurk
        if (
            askpass_prompted
            and askpass_path is not None
            and Path(askpass_path).is_file()
        ):
            os.remove(askpass_path)



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shutil
from dataclasses import dataclass, field
from pathlib import Path
from tempfile import TemporaryDirectory

from gurk.lib.context import GurkContext, Logger, get_plugin_directories
from gurk.lib.utils import EDITABLE_INSTALL


@dataclass
class SafeFileHandler:
    registries: dict[str, Path] = field(init=False, default_factory=dict)

    def __enter__(self):
        # Store original registry states for cleanup
        for plugin_dir in get_plugin_directories():
            with TemporaryDirectory() as tmp:
                tmp_dir = Path(tmp)
            shutil.copytree(plugin_dir, tmp_dir, symlinks=True)
            self.registries[plugin_dir.as_posix()] = tmp_dir

    def __exit__(self, exc_type, exc, tb):
        # Restore original registries
        for plugin_dir, tmp_dir in self.registries.items():
            if Path(plugin_dir).exists():
                shutil.rmtree(plugin_dir)
            shutil.copytree(tmp_dir, plugin_dir, symlinks=True)

        # Propagate exceptions
        return False


def main(argv):
    with GurkContext(
        logger=Logger(verbose=False, non_interactive=False, store_logs=False),
        writable=False,
    ) as ctx:
        # Check that pytest is installed
        try:
            import pytest
        except ImportError:
            ctx.logger.fatal(
                "'pytest' is not installed. Please install this package with the "
                "'dev' extras to use this command via: 'pipx install -e .[dev]'"
            )

        # Check that the package is installed in editable mode
        if not EDITABLE_INSTALL:
            ctx.logger.fatal(
                "Pytests can only be run with an editable installation of "
                "'gurk'. Please install this package with the 'dev' extras "
                "to use this command via: 'pipx install -e .[dev]'"
            )

    # Run pytests with a safe file handler to prevent any modifications to plugin registries during testing
    with (
        SafeFileHandler(),
        GurkContext(
            logger=None,
            writable=False,
        ) as ctx,
    ):
        # Run pytests
        raise SystemExit(pytest.main(argv))



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Literal, NotRequired, TypedDict, overload

from gurk.lib.shared.dicts import pprint_dict, pprint_typed_dict
from gurk.lib.shared.plugins import (
    PluginRegistryEntry,
    PluginSpecification,
    ResolvedPluginRegistry,
    ResolvedZippedRegistry,
)
from gurk.lib.shared.remotes import GitQuery, extract_url
from gurk.lib.utils import PathLike, full_isinstance, overlay_dicts, typecheck

from ..logger import get_logger
from .registry_manager import _get_registries
from .registry_utils import (
    _expand_registry_path,
    _filter_by_registries,
    _get_registry_files,
    _is_entry_valid,
    _zip_registry_files,
)


@overload
def get_registry_files(
    *,
    home_registry: Literal[True] = ...,
    package_registry: Literal[False] = ...,
) -> Path: ...


@overload
def get_registry_files(
    *,
    home_registry: Literal[False] = ...,
    package_registry: Literal[True] = ...,
) -> Path: ...


@overload
def get_registry_files(
    *,
    home_registry: Literal[True] = ...,
    package_registry: Literal[True] = ...,
) -> tuple[Path, Path]: ...


@typecheck
def get_registry_files(
    *, home_registry: bool, package_registry: bool
) -> tuple[Path, Path] | Path:
    """
    Get the registry files corresponding to the requested registries.

    :param home_registry: Whether to include the home plugin registry file
    :type home_registry: bool
    :param package_registry: Whether to include the package plugin registry file
    :type package_registry: bool
    :return: Tuple of plugin registry files (home, package), depending on the input
    :rtype: tuple[Path, Path] | Path
    """
    return _filter_by_registries(
        _get_registry_files(),
        home_registry=home_registry,
        package_registry=package_registry,
    )


@overload
def get_registries(
    *,
    home_registry: Literal[True] = ...,
    package_registry: Literal[False] = ...,
    combine: Literal[False] = ...,
) -> ResolvedPluginRegistry: ...


@overload
def get_registries(
    *,
    home_registry: Literal[False] = ...,
    package_registry: Literal[True] = ...,
    combine: Literal[False] = ...,
) -> ResolvedPluginRegistry: ...


@overload
def get_registries(
    *,
    home_registry: Literal[True] = ...,
    package_registry: Literal[True] = ...,
    combine: Literal[False] = ...,
) -> tuple[ResolvedPluginRegistry, ResolvedPluginRegistry]: ...


@overload
def get_registries(
    *,
    home_registry: Literal[True] = ...,
    package_registry: Literal[True] = ...,
    combine: Literal[True] = ...,
) -> ResolvedPluginRegistry: ...


@typecheck
def get_registries(
    *, home_registry: bool, package_registry: bool, combine: bool = False
) -> tuple[ResolvedPluginRegistry, ...] | ResolvedPluginRegistry:
    """
    Get the currently active registrator's plugin registries.

    :param home_registry: Whether to include the home plugin registry
    :type home_registry: bool
    :param package_registry: Whether to include the package plugin registry
    :type package_registry: bool
    :param combine: Whether to combine the home and package plugin registries into one, with the home registry prioritized
    :type combine: bool
    :return: Tuple of plugin registries (home, package), or a single (combined) registry, depending on the input
    :rtype: tuple[ResolvedPluginRegistry, ...] | ResolvedPluginRegistry
    :raises ValueError: If neither home_registry nor package_registry is True, or if combine is True while not both home_registry and package_registry are True
    :raises RuntimeError: If no registrator is initialized
    """
    # Validate input
    if not home_registry and not package_registry:
        raise ValueError(
            "At least one of home_registry or package_registry must be True"
        )
    if combine and not (home_registry and package_registry):
        raise ValueError(
            "Combine can only be True if both home_registry and package_registry are True"
        )

    # Filter registries based on input
    filtered = _filter_by_registries(
        _get_registries(),
        home_registry=home_registry,
        package_registry=package_registry,
        dcopy=True,
    )

    if combine:
        return overlay_dicts(filtered)
    else:
        return filtered


@overload
def _get_zipped_registries(
    *,
    home_registry: Literal[True] = ...,
    package_registry: Literal[False] = ...,
) -> ResolvedZippedRegistry: ...


@overload
def _get_zipped_registries(
    *,
    home_registry: Literal[False] = ...,
    package_registry: Literal[True] = ...,
) -> ResolvedZippedRegistry: ...


@overload
def _get_zipped_registries(
    *,
    home_registry: Literal[True] = ...,
    package_registry: Literal[True] = ...,
) -> tuple[ResolvedZippedRegistry, ResolvedZippedRegistry]: ...


@typecheck
def _get_zipped_registries(
    *, home_registry: bool, package_registry: bool
) -> (
    tuple[ResolvedZippedRegistry, ResolvedZippedRegistry]
    | ResolvedZippedRegistry
):
    """
    Get a tuple of tuples of plugin registry files and their corresponding registries, with the home one first.

    :param home_registry: Whether to include the home plugin registry
    :type home_registry: bool
    :param package_registry: Whether to include the package plugin registry
    :type package_registry: bool
    :return: Tuple of tuples of plugin registry files and their corresponding registries (home, package), depending on the input
    :rtype: tuple[ResolvedZippedRegistry, ResolvedZippedRegistry] | ResolvedZippedRegistry
    """
    _registries = get_registries(home_registry=True, package_registry=True)
    return _filter_by_registries(
        _zip_registry_files(_registries),
        home_registry=home_registry,
        package_registry=package_registry,
    )


@typecheck
def _get_plugin_registration(
    plugin: PluginSpecification,
    *,
    package_registry: bool,
    require_local: bool,
) -> ResolvedPluginRegistry | None:
    """
    Get the registry entry of a plugin (path, remote) if it exists locally.

    :param plugin: Name, PathLike, or GitQuery of the plugin
    :type plugin: PluginSpecification
    :param package_registry: Whether to check the package plugin registry (True) or the home plugin registry (False)
    :type package_registry: bool
    :param require_local: Whether to only return entries with a local path
    :type require_local: bool
    :return: Registry entry if the plugin exists locally, None otherwise
    :rtype: ResolvedPluginRegistry | None
    :raises RuntimeError: If no registrator is initialized
    """
    # Get the plugin registry file and registry
    registry_file, registry = _get_zipped_registries(
        home_registry=not package_registry, package_registry=package_registry
    )

    # Access plugin entry
    name_via_remote = next(
        (
            k
            for k, v in registry.items()
            if v.get("remote")
            and extract_url(str(plugin)) == extract_url(v["remote"])
        ),
        None,
    )
    name_via_local = next(
        (
            k
            for k, v in registry.items()
            if v.get("local") is not None
            and _expand_registry_path(registry_file, v["local"])
            == registry_file.parent / Path(plugin).expanduser()
        ),
        None,
    )
    if str(plugin) in registry:
        # Access plugin by name
        name = str(plugin)
    elif name_via_remote:
        # Access plugin by remote
        name = name_via_remote
    elif name_via_local:
        # Access plugin by local path
        name = name_via_local
    else:
        # Plugin not found
        return None
    entry = registry[name]

    # Validate that it has a valid local path (if required)
    if require_local and (
        entry["local"] is None
        or not _expand_registry_path(registry_file, entry["local"]).is_dir()
    ):
        return None

    return {name: entry}


def get_available_plugin_names() -> set[str]:
    """
    Get the names of all available local plugins, extracted from the combined registry keys.

    :return: List of available local plugin names
    :rtype: list[str]
    """
    combined_registry = get_registries(
        home_registry=True, package_registry=True, combine=True
    )
    return set(combined_registry.keys())


@typecheck
def get_plugin_registration(
    plugin: PluginSpecification,
    *,
    home_registry: bool,
    package_registry: bool,
    require_local: bool = True,
) -> ResolvedPluginRegistry | None:
    """
    Get the registry entry of a plugin (path, remote) if it exists locally.

    :param plugin: Name, PathLike, or GitQuery of the plugin
    :type plugin: PluginSpecification
    :param package_registry: Whether to check the package plugin registry (True) or the home plugin registry (False)
    :type package_registry: bool
    :param require_local: Whether to only return entries with a local path
    :type require_local: bool
    :return: Registry entry if the plugin exists locally, None otherwise
    :rtype: ResolvedPluginRegistry | None
    :raises RuntimeError: If no registrator is initialized
    """
    # Collect registrations
    registrations = tuple(
        _get_plugin_registration(
            plugin,
            package_registry=pr,
            require_local=require_local,
        )
        for ind, pr in enumerate((False, True))
        if (home_registry, package_registry)[ind]
    )
    registrations = tuple(
        r for r in registrations if r
    )  # Filter out None results

    # Logging
    logger = get_logger()
    if not registrations:
        return None
    elif len(registrations) > 1:
        logger.debug(
            f"WARNING: Multiple registry entries found for plugin '{plugin}'. Using the home one."
        )

    return registrations[0]


@typecheck
def is_plugin_registered(
    plugin: PluginSpecification,
    *,
    home_registry: bool,
    package_registry: bool,
    require_local: bool = True,
) -> bool:
    """
    Check if a plugin is registered locally in the currently active registrator's plugin registry.

    :param plugin: Name, PathLike, or GitQuery of the plugin
    :type plugin: PluginSpecification
    :param home_registry: Whether to check the home plugin registry (takes precedence over the package registry if both are True)
    :type home_registry: bool
    :param package_registry: Whether to check the package plugin registry
    :type package_registry: bool
    :param require_local: Whether to only consider entries with a local path as registered
    :type require_local: bool
    :return: Whether the plugin is registered locally
    :rtype: bool
    :raises RuntimeError: If no registrator is initialized
    """
    registration = get_plugin_registration(
        plugin,
        home_registry=home_registry,
        package_registry=package_registry,
        require_local=require_local,
    )
    return registration is not None


class LocalPluginRegistryEntry(TypedDict):
    # fmt: off
    local:   None | PathLike
    remote:  NotRequired[None | GitQuery]
    # fmt: on


class RemotePluginRegistryEntry(TypedDict):
    # fmt: off
    local:   NotRequired[None | PathLike]
    remote:  None | GitQuery
    # fmt: on


@typecheck
def update_registry(
    plugin_name: str,
    entry: LocalPluginRegistryEntry | RemotePluginRegistryEntry | None,
    *,
    package_registry: bool | None = None,
    infer_local: bool = False,
    exist_ok: bool = True,
) -> bool:
    """
    Update the currently active registrator's plugin registry with a new entry.

    :param plugin_name: Name of the plugin to update
    :type plugin_name: str
    :param entry: Plugin registry entry to add or update, or None to remove the entry
    :type entry: LocalPluginRegistryEntry | RemotePluginRegistryEntry | None
    :param package_registry: Whether to update the package plugin registry (if False, updates the home registry). If None, tries to determine on where the plugin is already registered and returns False if it cannot be determined (i.e. when adding a new plugin)
    :type package_registry: bool | None
    :param infer_local: Whether to infer the "local" value of the entry (does not apply when already specified or when removing an entry)
    :type infer_local: bool
    :param exist_ok: Whether to allow updating an existing entry (if False, raises an error if the entry already exists)
    :type exist_ok: bool
    :return: True if the registry was updated, False otherwise
    :rtype: bool
    :raises RuntimeError: If no registrator is initialized
    """
    # Get logger
    logger = get_logger()

    def debug_error(message: str) -> None:
        logger.debug(
            f"ERROR: Invalid 'entry' value passed for updating plugin '{plugin_name}': {message}"
        )

    # Validate input
    if entry is not None:
        if not (
            full_isinstance(entry, LocalPluginRegistryEntry)
            or full_isinstance(entry, RemotePluginRegistryEntry)
        ):
            debug_error(
                f"'entry' has an invalid structure. Expected:\n"
                f"{pprint_typed_dict(PluginRegistryEntry, indent=2, as_str=True)}"
                f"but got:\n{pprint_dict(entry, indent=2, as_str=True)}"
            )
            return False
        if not (
            infer_local or entry.get("local") is not None
        ) and not entry.get("remote"):
            debug_error("'entry' must have at least a local or remote value.")
            return False

    # Get registries
    registries = _get_registries()

    # Determine registry
    if package_registry is None:
        if is_plugin_registered(
            plugin_name,
            home_registry=True,
            package_registry=False,
            require_local=False,
        ):
            package_registry = False
        elif is_plugin_registered(
            plugin_name,
            home_registry=False,
            package_registry=True,
            require_local=False,
        ):
            package_registry = True
        else:
            logger.debug(
                "Could not determine registry, since it is registered in "
                "neither registry. Defaulting to home registry for new entry."
            )
            package_registry = False
    registry_index = int(package_registry)
    registry_str = "package" if package_registry else "home"
    registry_file = _get_registry_files()[registry_index]

    # Update registry
    if (
        entry is not None
        and plugin_name in registries[registry_index]
        and not exist_ok
    ):
        debug_error(f"Entry already exists in the {registry_str} registry")
        return False
    else:
        if entry is None:
            # Remove registration
            if package_registry and plugin_name in registries[registry_index]:
                # Can't fully remove package plugins' registration
                if registries[registry_index][plugin_name].get("remote"):
                    registries[registry_index][plugin_name]["local"] = None
                    return True
                else:
                    debug_error(
                        f"Cannot remove plugin '{plugin_name}' from the {registry_str} registry "
                        "since it does not have a remote. If you still want to remove it, please "
                        f"remove the plugin directory ({registry_file.parent / plugin_name}) "
                        f"and its entry from the registry file ({registry_file}) manually."
                    )
                    return False
            else:
                registries[registry_index].pop(plugin_name, None)
                return True
        else:
            # Infer local path if applicable
            if infer_local and entry.get("local") is None:
                entry["local"] = registry_file.parent / plugin_name
            elif entry.get("local") is not None:
                entry["local"] = Path(entry["local"]).expanduser()

            # Add/update registration
            overlayed = overlay_dicts(
                [registries[registry_index].get(plugin_name, {}), entry]
            )
            if _is_entry_valid(
                plugin_name, overlayed, registry_file, check_local=False
            ):
                registries[registry_index][plugin_name] = overlayed
                return True
            else:
                debug_error(
                    f"Cannot add/update registration in {registry_str} "
                    "registry, as the resulting entry would be invalid/incomplete."
                )
                return False



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess
from dataclasses import dataclass, field
from pathlib import Path
from textwrap import dedent

from gurk.lib.context import GurkContext, Logger, get_logger
from gurk.lib.core.plugins import DefaultNamespace, GurkArgumentParser
from gurk.lib.shared.printers import newline, padded_print, richprint
from gurk.lib.shared.system_info import get_manufacturer
from gurk.lib.utils import SETUP_DONE_FILE


@dataclass
class SSHKeysManager:
    """
    Manage SSH keys for the user.
    """

    ssh_directory: Path = field(
        init=False, default=Path("~/.ssh").expanduser()
    )

    def keys_exist(self) -> bool:
        """
        Check if any SSH keys are already added to the ssh-agent.

        :return: True if keys exist, False otherwise.
        :rtype: bool
        """
        try:
            output = (
                subprocess.check_output(
                    ["ssh-add", "-l"], stderr=subprocess.DEVNULL
                )
                .decode()
                .strip()
            )
            if "The agent has no identities" in output or not output:
                return False
            return True
        except subprocess.CalledProcessError:
            # ssh-agent may not be running
            return False

    def setup_key(self) -> None:
        # Get logger
        logger = get_logger()
        padded_print("New SSH Key", "cyan", 64)

        # Get key name
        padded_print("SSH Key Name", "yellow", 32, top=False, bottom=False)
        while True:
            key_name = logger.ask(
                "Enter a name for your SSH key (e.g. id_ed25519)"
            ).strip()
            if key_name:
                curr_ssh_key = self.ssh_directory / key_name
                ssh_key_pub_path = curr_ssh_key.with_suffix(".pub")
                if curr_ssh_key.is_file():
                    logger.warning(
                        f"Key '{key_name}' already exists. Please choose a different name.\n"
                    )
                    continue
                break
            else:
                logger.warning("Key name cannot be empty. Please try again.\n")

        # Get key password
        padded_print("SSH Key Password", "yellow", 32, top=False, bottom=False)
        while True:
            password = logger.ask(
                "Enter a password for the SSH key (can be empty)",
                password=True,
            )
            password_confirm = logger.ask(
                "Confirm the password", password=True
            )
            if password == password_confirm:
                break
            logger.warning("Passwords do not match. Try again.")

        # Create key
        os.makedirs(self.ssh_directory, exist_ok=True)
        subprocess.run(
            [
                "ssh-keygen",
                "-t",
                "ed25519",
                "-f",
                curr_ssh_key,
                "-N",
                password,
            ],
            capture_output=True,
        )
        subprocess.run(["eval", "$(ssh-agent -s)"], shell=True)
        subprocess.run(["ssh-add", curr_ssh_key])

        # Prompt the user to upload the public SSH key
        padded_print("SSH Key Upload", "yellow", 32, top=False, bottom=False)
        richprint(
            f"Please upload the public key ({ssh_key_pub_path}) to your account settings (GitHub, GitLab, etc.). Public key:"
        )
        with open(ssh_key_pub_path) as f:
            richprint(f.read().strip(), color="green")
        input("After uploading your key, press anything to continue...")

    def setup_keys(self) -> None:
        """
        Set up SSH keys by prompting the user for input.
        """
        # Get logger
        logger = get_logger()
        while True:
            self.setup_key()
            if not logger.prompt_bool(
                "Would you like to create another SSH key?"
            ):
                break


@dataclass
class GitCredentialsManager:
    """
    Manage Git user credentials (name and email).
    """

    # fmt: off
    user_name:  str = field(default="")
    user_email: str = field(default="")
    # fmt: on

    def credentials_exist(self) -> bool:
        """
        Check if git user name and email are already set.

        :return: True if both user name and email are set, False otherwise
        :rtype: bool
        """
        # Get username
        try:
            self.user_name = (
                subprocess.check_output(
                    ["git", "config", "--global", "user.name"]
                )
                .decode()
                .strip()
            )
        except subprocess.CalledProcessError:
            self.user_name = ""

        # Get user email
        try:
            self.user_email = (
                subprocess.check_output(
                    ["git", "config", "--global", "user.email"]
                )
                .decode()
                .strip()
            )
        except subprocess.CalledProcessError:
            self.user_email = ""

        # Return whether both are set
        return bool(self.user_name and self.user_email)

    def setup_credentials(self) -> None:
        """Set up git user name and email."""
        # Get logger
        logger = get_logger()
        padded_print("Git User Info", "cyan", 64)

        # Prompt for username
        padded_print("Git User Name", "yellow", 32, top=False, bottom=False)
        while True:
            self.user_name = logger.ask("Enter your Git username").strip()
            if not self.user_name:
                logger.warning("Username cannot be empty. Please try again.")
                continue
            break

        # Prompt for email
        padded_print("Git User Email", "yellow", 32, top=False, bottom=False)
        while True:
            self.user_email = logger.ask("Enter your Git email").strip()
            if not self.user_email:
                logger.warning("Email cannot be empty. Please try again.")
                continue
            elif (
                "@" not in self.user_email
                or "." not in self.user_email.split("@")[-1]
            ):
                logger.warning("Invalid email format. Please try again.")
                continue
            break

        # Set credentials
        subprocess.run(
            ["git", "config", "--global", "user.name", self.user_name]
        )
        subprocess.run(
            ["git", "config", "--global", "user.email", self.user_email]
        )


def print_secure_boot_steps() -> None:
    """
    Print steps to disable Secure Boot in UEFI/BIOS.
    """
    # Get logger
    logger = get_logger()
    padded_print("Disable Secure Boot Steps", "cyan", 64)

    # Table of common manufacturers → probable keys
    key_table = {
        "acer": ["F2", "Del", "F12"],
        "asus": ["F2", "Del"],
        "dell": ["F2", "F12"],
        "hp": ["Esc", "F10", "F2"],
        "lenovo": ["F1", "F2", "Novo button"],
        "msi": ["Del", "F11"],
        "gigabyte": ["Del"],
        "asrock": ["Del", "F2"],
        "toshiba": ["F2", "Esc"],
        "samsung": ["F2"],
        "sony": ["F2", "Assist button"],
        "microsoft": ["Volume Up"],
        "system76": ["F2", "Del"],
        "purism": ["Del", "Esc"],
    }

    # Find best match
    manufacturer = get_manufacturer()
    matches = [
        (name, keys)
        for name, keys in key_table.items()
        if name in manufacturer
    ]
    if not matches:
        all_keys_str = "Esc, Del, F1, F2, F10, F12"
    else:
        all_keys_str = ", ".join([key for _, keys in matches for key in keys])

    # Print steps
    richprint(dedent(f"""\
            1. Reboot your computer. During the initial boot screen, repeatedly press one of the following keys to enter the UEFI/BIOS setup: {all_keys_str}
            2. Navigate to the 'Security' or 'Boot' tab using the arrow keys, locate the 'Secure Boot' option and disable it.
            3. Save your changes and exit the UEFI/BIOS setup - Your computer will reboot with Secure Boot disabled.
        """))


class SetupNamespace(DefaultNamespace):
    # fmt: off
    ssh_keys:            bool
    git_credentials:     bool
    disable_secure_boot: bool
    # fmt: on


def main(argv, prog, description):
    parser = GurkArgumentParser[SetupNamespace](
        prog=prog,
        description=description,
        add_verbose_arg=False,
        add_non_interactive_arg=False,
    )
    # Flags to short-circuit specific pre-setup tasks
    parser.add_argument(
        "-s",
        "--ssh-keys",
        action="store_true",
        help="Set up SSH keys without prompt",
    )
    parser.add_argument(
        "-g",
        "--git-credentials",
        action="store_true",
        help="Set up Git Credentials (username, email) without prompt",
    )
    parser.add_argument(
        "-d",
        "--disable-secure-boot",
        action="store_true",
        help="Print steps to disable Secure Boot in UEFI/BIOS",
    )
    args = parser.parse_args(argv)

    # If none are enabled, enable all
    flags = vars(args)
    if not any(flags.values()):
        for flag in flags:
            flags[flag] = True

    # Execute without writing to plugins
    with GurkContext(
        logger=Logger(verbose=False, non_interactive=False, store_logs=False),
        writable=False,
    ) as ctx:
        # Set up SSH keys
        ssh_keys_manager = SSHKeysManager()
        ssh_keys_exist = ssh_keys_manager.keys_exist()
        if args.ssh_keys:
            if (
                not ssh_keys_exist
                and ctx.logger.prompt_bool(
                    "No SSH keys detected. Would you like to create SSH keys?"
                )
            ) or (
                ssh_keys_exist
                and ctx.logger.prompt_bool(
                    "Existing SSH keys detected. Would you still like to create new ones?"
                )
            ):
                ssh_keys_manager.setup_keys()
            newline()

        # Set up Git Credentials
        git_credentials_manager = GitCredentialsManager()
        git_credentials_exist = git_credentials_manager.credentials_exist()
        if args.git_credentials:
            if (
                not git_credentials_exist
                and ctx.logger.prompt_bool(
                    "Git user name/email not set. Would you like to set them up?"
                )
            ) or (
                git_credentials_exist
                and ctx.logger.prompt_bool(
                    f"Git user name/email already set (to '{git_credentials_manager.user_name}' resp. '{git_credentials_manager.user_email}'). Would you like to update them?"
                )
            ):
                git_credentials_manager.setup_credentials()
            newline()

        # Print Secure Boot disabling steps
        if args.disable_secure_boot and ctx.logger.prompt_bool(
            "Would you like to see the steps to disable Secure Boot in UEFI/BIOS?"
        ):
            print_secure_boot_steps()

        # Mark setup as done
        if not SETUP_DONE_FILE.is_file():
            SETUP_DONE_FILE.parent.mkdir(parents=True, exist_ok=True)
            SETUP_DONE_FILE.touch()



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from gurk.lib.context import GurkContext
from gurk.lib.core.plugins import (
    GurkArgumentParser,
    TaskParserNamespace,
    get_resolved_plugin_manifest,
)
from gurk.lib.utils import ListOrTuple, typecheck


@typecheck
def _extend_task_arguments(parser: GurkArgumentParser, task_name: str) -> None:
    """
    Extend the parser with task-specific arguments defined in a plugin, if any.

    :param parser: Argument parser to extend with task-specific arguments
    :type parser: GurkArgumentParser
    :param task_name: Full name of a task in the form 'plugin_name/task_name'
    :type task_name: str
    :raises ValueError: If the plugin YAML could not be loaded or if the task arguments could not be found in the plugin manifest
    """
    plugin = task_name.split("/", 1)[0]
    plugin_manifest = get_resolved_plugin_manifest(plugin)
    if plugin_manifest is None:
        raise ValueError(f"Plugin '{plugin}' could not be loaded")

    try:
        task_args = plugin_manifest["tasks"][task_name]["args"]
        parser.extend_arguments(task_args)
    except KeyError as e:
        raise ValueError(
            f"Key 'tasks'/'{task_name}'/'args' not found "
            f"in plugin '{plugin}' YAML. Broken link: {e}"
        )


@typecheck
def parse_task_args(
    args: ListOrTuple[str],
) -> TaskParserNamespace:
    """
    Parse command-line arguments and return system info, config info, and remaining args.

    :param args: Arguments passed to the task, including task name at the start
    :type args: ListOrTuple[str]
    :return: Parsed system info, config file path, force flag and remaining arguments
    :rtype: TaskParserNamespace
    """
    task_name, *remaining_args = args
    parser = GurkArgumentParser[TaskParserNamespace](
        add_verbose_arg=False,
        add_non_interactive_arg=False,
        add_force_arg=True,
        add_task_args=True,
    )
    with GurkContext(logger=None, writable=False):
        _extend_task_arguments(parser, task_name)

    return parser.parse_args(remaining_args)



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from gurk.lib.context import GurkContext, Logger
from gurk.lib.core.plugins import (
    DefaultNamespace,
    GurkArgumentParser,
    install_plugin,
)
from gurk.lib.shared.remotes import is_git_installed


class PullNamespace(DefaultNamespace):
    sources: list[str]
    replace: bool


def main(argv, prog, description):
    parser = GurkArgumentParser[PullNamespace](
        prog=prog, description=description
    )
    group = parser.add_required_group()
    group.add_argument(
        "sources",
        type=str,
        nargs="+",
        help="Local paths or GitQueries of the plugin to install from",
    )
    parser.add_argument(
        "-r",
        "--replace",
        action="store_true",
        help="Replace existing plugins if they already exist (in any form)",
    )
    args = parser.parse_args(argv)

    # Execute with writing to plugins
    with GurkContext(
        logger=Logger(
            verbose=args.verbose,
            non_interactive=args.non_interactive,
            description="Pulling plugins",
        ),
        writable=True,
    ) as ctx:
        # Check that git is installed
        if not is_git_installed():
            ctx.logger.fatal(
                "Git is not installed or not available in PATH."
                "Please install it via 'sudo apt install git'"
            )

        # (Re)install specified plugins
        for source in args.sources:
            install_plugin(source, reinstall=args.replace)

        ctx.logger.done("Plugin pulling complete.")



import os
import subprocess
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import NotRequired, TypedDict

import commentjson

from gurk import dump_yaml, log_step, parse_task_args, run_script_function


def install_conda_environments(*args: list[str]) -> None:
    """
    Install packages into Conda environments (no custom env directory).
    """
    # Parse task args
    task_args = parse_task_args(args)

    # Typing helper classes
    class CondaEnv(TypedDict):
        # fmt: off
        type:           str                   # "conda", "mamba"
        conda_packages: dict[str, list[str]]  # package-name -> [channels]
        pip_packages:   NotRequired[list[str]]
        # fmt: on

    # Get conda environments info
    conda_envs: dict[str, CondaEnv] = commentjson.load(
        task_args.config_file.open("r", encoding="utf-8")
    )
    if not conda_envs:
        log_step(
            "Skipping installation of conda environments, as no environments are specified",
        )
        return

    # Check if conda types are installed
    conda_exe = {"conda": None, "mamba": None}
    for conda_type in conda_exe.keys():
        result = run_script_function(
            script=Path(__file__).parent / "checks.bash",
            function=f"check_install_{conda_type}",
            run=True,
            capture_output=True,
        )
        if result.returncode == 0:
            conda_exe[conda_type] = result.stdout.strip()

    def check_env_type(env_type: str | None) -> bool:
        """Check if conda environment type field is valid."""
        if env_type is None:
            log_step(
                f"No environment type specified for '{env_name}' - Skipping",
                warning=True,
            )
            return False
        elif env_type not in conda_exe.keys():
            log_step(
                f"Unsupported environment type '{env_type}' for '{env_name}' - Skipping",
                warning=True,
            )
            return False

        for conda_type, exe in conda_exe.items():
            if env_type == conda_type and exe is None:
                log_step(
                    f"'{env_type}' is not installed, cannot create environment '{env_name}' - Skipping",
                    warning=True,
                )
                return False

        return True

    # (STEP) Creating conda environments
    for env_name, env_spec in conda_envs.items():
        # Get and check conda environment type
        env_type = env_spec.get("type", None)
        if not check_env_type(env_type):
            continue

        # Get desired packages
        conda_packages = env_spec.get("conda_packages", {})
        pip_packages = env_spec.get("pip_packages", [])
        if not conda_packages and not pip_packages:
            log_step(
                f"Skipping installation of conda environment '{env_name}', as no packages are specified",
                warning=True,
            )
            continue

        # Get channels
        channels = env_spec.get("channels", [])

        # Environment config file
        env_spec = {
            "name": env_name,
            "channels": channels,
            "dependencies": conda_packages,
        }
        if pip_packages:
            env_spec["dependencies"].append("pip")
            env_spec["dependencies"].append({"pip": pip_packages})

        env_file = NamedTemporaryFile(delete=False, suffix=".yaml").name
        dump_yaml(env_file, env_spec)

        # Executable command
        conda_cmd = [
            conda_exe[env_type],
            "env",
            "create",
            "-y",
            "-f",
            env_file,
        ]

        # Check if environment already exists
        check_cmd = [
            conda_exe[env_type],
            "run",
            "-n",
            env_name,
            "echo",
            "Environment exists",
        ]
        result = subprocess.run(
            check_cmd,
            capture_output=True,
            text=True,
        )
        if (
            result.returncode == 0
            and not task_args.conda_update_environments
            and not task_args.force
        ):
            log_step(
                f"Environment '{env_name}' already exists - Skipping creation",
                warning=True,
            )
            continue

        # Handle --update flag
        if task_args.conda_update_environments:
            conda_cmd = ["update" if x == "create" else x for x in conda_cmd]
        else:
            if task_args.force:
                result = subprocess.run(
                    [conda_exe[env_type], "env", "remove", "-n", env_name],
                    capture_output=True,
                    text=True,
                )
                if not result.returncode == 0:
                    log_step(
                        f"Failed to remove existing environment '{env_name}' - Skipping creation",
                        warning=True,
                    )
                    continue
            else:
                log_step(
                    f"Environment '{env_name}' already exists - Skipping creation",
                    warning=True,
                )
                continue

        # Create environment
        log_step(
            f"Creating environment '{env_name}' with {env_type}...",
        )
        result = subprocess.run(conda_cmd)
        if result.returncode != 0:
            log_step(
                f"Failed to create environment '{env_name}'", warning=True
            )
        else:
            log_step(f"Successfully created environment '{env_name}'")

        # Cleanup
        os.remove(env_file)



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

from gurk.lib.context import GurkContext, Logger, get_registries
from gurk.lib.context.registry import (
    get_plugin_registration,
    is_plugin_registered,
)
from gurk.lib.core.plugins import (
    DefaultNamespace,
    GurkArgumentParser,
    is_plugin_installed,
    upgrade_plugin,
)
from gurk.lib.shared.remotes import extract_url, is_git_installed, is_git_repo


class UpgradeNamespace(DefaultNamespace):
    plugins: list[str]
    exclude: list[str] | None


def main(argv, prog, description):
    parser = GurkArgumentParser[UpgradeNamespace](
        prog=prog, description=description
    )
    group = parser.add_required_group()
    group.add_argument(
        "plugins",
        type=str,
        nargs="*",
        help="Names of the registered plugins to upgrade. If empty, upgrade all local plugins",
    )
    parser.add_argument(
        "-e",
        "--exclude",
        type=str,
        nargs="+",
        help="PluginSpecifications (name or remote) to exclude from upgrade when upgrading all plugins",
    )
    args = parser.parse_args(argv)

    # Execute with writing to plugins
    with GurkContext(
        logger=Logger(
            verbose=args.verbose,
            non_interactive=args.non_interactive,
            description="Upgrading plugins",
        ),
        writable=True,
    ) as ctx:
        # Check that git is installed
        if not is_git_installed():
            ctx.logger.fatal(
                "Git is not installed or not available in PATH."
                "Please install it via 'sudo apt install git'"
            )

        # Parse plugin specifications to upgrade
        if args.plugins:
            plugins = args.plugins
            for plugin in plugins:
                # Only allow plugin names
                if Path(plugin).exists() or is_git_repo(plugin):
                    ctx.logger.error(
                        f"Invalid plugin specification '{plugin}' given "
                        f"for upgrade. Only plugin names are allowed."
                    )
                    plugins.remove(plugin)
        else:
            # Get all local plugins to upgrade if none specified
            combined_registry = get_registries(
                home_registry=True, package_registry=True, combine=True
            )
            plugins = combined_registry.keys()  # All plugin names

        # Check 'exclude' plugins if specified
        normalized_exclude = set()
        for exclude in args.exclude or []:
            # Don't allow local paths
            if Path(exclude).exists():
                ctx.logger.error(
                    f"Invalid plugin specification '{exclude}' given for exclusion. "
                    f"Only plugin names or remotes are allowed. Skipping..."
                )
                continue

            # Check that the plugin to exclude exists
            if not is_plugin_installed(exclude, require_venv=False):
                ctx.logger.warning(
                    f"Excluded plugin '{exclude}' is not validly installed. Ignoring..."
                )
                continue
            else:
                normalized_exclude.add(extract_url(exclude))

        for plugin in plugins:
            # Check if plugin is registered
            if not is_plugin_registered(
                plugin,
                home_registry=True,
                package_registry=True,
                require_local=False,
            ):
                ctx.logger.error(
                    f"Plugin '{plugin}' is not registered. Skipping upgrade..."
                )
                continue

            # Get available plugin data from registration
            plugin_registration = get_plugin_registration(
                plugin,
                home_registry=True,
                package_registry=True,
                require_local=False,
            )
            plugin_name, plugin_entry = next(iter(plugin_registration.items()))
            plugin_local = (
                (
                    plugin_entry["local"].as_posix()
                    if isinstance(plugin_entry["local"], Path)
                    else plugin_entry["local"]
                ),
            )
            plugin_remote = plugin_entry["remote"]

            # Skip local-only plugins
            if not plugin_remote:
                ctx.logger.error(
                    f"Plugin '{plugin}' is local-only and has no remote. Skipping upgrade..."
                )
                continue

            # Exclude specified plugins
            if normalized_exclude & {
                plugin_name,
                plugin_local,
                extract_url(plugin_remote),
            }:
                ctx.logger.debug(f"Excluding plugin '{plugin}' from upgrade.")
                continue

            # Upgrade plugin from remotes
            upgrade_plugin(plugin)

        ctx.logger.done("Plugin upgrades complete.")



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from pathlib import Path
from typing import Literal, TypeVar, get_type_hints, overload

from gurk.lib.shared.plugins import (
    PluginRegistry,
    ResolvedPluginRegistry,
    ResolvedPluginRegistryEntry,
    ResolvedZippedRegistry,
    ZippedRegistry,
)
from gurk.lib.shared.remotes import GitQueryDict, is_git_repo, parse_git_query
from gurk.lib.utils import (
    PACKAGE_HOME_PATH,
    PACKAGE_SRC_PATH,
    PathLike,
    full_isinstance,
    typecheck,
)

from ..logger import get_logger


@typecheck
def _deepcopy_tuple(tup: tuple) -> tuple:
    """
    Deepcopy a tuple by deepcopying each item and returning a new tuple.

    :param tup: Tuple to deepcopy
    :type tup: tuple
    :return: Deepcopied tuple
    :rtype: tuple
    """
    return tuple(deepcopy(item) for item in tup)


@overload
def get_plugin_directories(
    *,
    home_registry: Literal[True] = ...,
    package_registry: Literal[True] = ...,
) -> tuple[Path, Path]: ...


@overload
def get_plugin_directories(
    *,
    home_registry: Literal[True] = ...,
    package_registry: Literal[False] = ...,
) -> Path: ...


@overload
def get_plugin_directories(
    *,
    home_registry: Literal[False] = ...,
    package_registry: Literal[True] = ...,
) -> Path: ...


@typecheck
def get_plugin_directories(
    *, home_registry: bool = True, package_registry: bool = True
) -> tuple[Path, Path] | Path:
    """
    Get a tuple of plugin directories, with the home one first.

    :param home_registry: Whether to include the home plugin directory
    :type home_registry: bool
    :param package_registry: Whether to include the package plugin directory
    :type package_registry: bool
    :return: Tuple of plugin directories (home, package), depending on the input
    :rtype: tuple[Path, Path] | Path
    :raises TypeError: If an expected plugin directory path exists but is not a directory
    """
    parent_paths = (PACKAGE_HOME_PATH, PACKAGE_SRC_PATH)

    possible_plugin_paths = [p / "plugins" for p in parent_paths]
    for p in possible_plugin_paths:
        if p.is_file():
            raise TypeError(
                f"Expected plugin directory at '{p.as_posix()}', but "
                "found a file. Please remove or rename this file."
            )
        p.mkdir(parents=True, exist_ok=True)

    return _filter_by_registries(
        tuple(possible_plugin_paths),
        home_registry=home_registry,
        package_registry=package_registry,
    )


def _get_registry_files() -> tuple[Path, Path]:
    """
    Get a tuple of plugin registries, with the home one first.

    :return: Tuple of plugin registries (home, package)
    :rtype: tuple[Path, Path]
    """
    plugin_registry_files = [
        p / "registry.yaml" for p in get_plugin_directories()
    ]
    for p in plugin_registry_files:
        p.touch(exist_ok=True)

    return tuple(plugin_registry_files)


@overload
def _zip_registry_files(
    registries: tuple[PluginRegistry, PluginRegistry],
) -> tuple[ZippedRegistry, ZippedRegistry]: ...


@overload
def _zip_registry_files(
    registries: tuple[ResolvedPluginRegistry, ResolvedPluginRegistry],
) -> tuple[ResolvedZippedRegistry, ResolvedZippedRegistry]: ...


def _zip_registry_files(
    registries: tuple[
        PluginRegistry | ResolvedPluginRegistry,
        PluginRegistry | ResolvedPluginRegistry,
    ],
) -> tuple[
    ZippedRegistry | ResolvedZippedRegistry,
    ZippedRegistry | ResolvedZippedRegistry,
]:
    """
    Zip the plugin registries with their corresponding registry files, with the home one first.

    :param registries: Tuple of plugin registries (home, package)
    :type registries: tuple[PluginRegistry | ResolvedPluginRegistry, PluginRegistry | ResolvedPluginRegistry]
    :return: Tuple of tuples of plugin registry files and their corresponding registries (home, package)
    :rtype: tuple[ZippedRegistry | ResolvedZippedRegistry, ZippedRegistry | ResolvedZippedRegistry]
    """
    return tuple(zip(_get_registry_files(), registries))


@typecheck
def _expand_registry_path(
    registry_file: PathLike, path: PathLike, collapse: bool = False
) -> Path:
    """
    Expand the "local" path of a plugin registry entry to an absolute path based on the registry file's location.

    :param registry_file: Path to the registry file
    :type registry_file: PathLike
    :param path: Path to expand
    :type path: PathLike
    :param collapse: Whether to collapse the path instead of expanding it
    :type collapse: bool
    :return: Expanded plugin registry entry
    :rtype: Path
    """
    registry_file = Path(registry_file).expanduser()
    if collapse:
        return Path(path).relative_to(registry_file.parent)
    else:
        return (registry_file.parent / path).expanduser().resolve()


@typecheck
def _is_entry_valid(
    name: str,
    entry: ResolvedPluginRegistryEntry,
    registry_file: Path,
    check_local: bool = True,
) -> bool:
    """
    Check if a plugin registry entry is valid.

    :param name: Name of the plugin
    :param entry: Plugin registry entry to validate
    :param registry_file: Path to the registry file
    :param check_local: Whether to check that the local path exists on disk (if specified).
    :return: Whether the entry is valid
    """
    checks: dict[str, bool] = {
        f"Plugin name '{name}' must be a string": isinstance(name, str),
        "Entry must conform to ResolvedPluginRegistryEntry schema": full_isinstance(
            entry, ResolvedPluginRegistryEntry
        ),
        "Entry must define at least one of 'local' or 'remote'": any(
            (entry.get("local") is not None, entry.get("remote") is not None)
        ),
        f"Local path {entry.get('local')} must exist if specified": (
            entry.get("local") is None
            or (
                not check_local
                or _expand_registry_path(
                    registry_file, entry["local"]
                ).is_dir()
            )
        ),
        f"Remote {entry.get('remote')} must be a valid git repository if specified": (
            entry.get("remote") is None or is_git_repo(entry["remote"])
        ),
        f"Remote {entry.get('remote')} must define a commit (and nothing else) if specified": (
            entry.get("remote") is None
            or (
                parse_git_query(entry["remote"])["commit"] is not None
                and all(
                    parse_git_query(entry["remote"])[f] is None
                    for f in get_type_hints(GitQueryDict)
                    if f not in ("url", "commit")
                )
            )
        ),
    }

    failed_checks = [msg for msg, passed in checks.items() if not passed]

    if failed_checks:
        failed_msgs = "\n- " + "\n- ".join(failed_checks)
        get_logger().debug(
            f"Invalid plugin registry entry '{name}' in "
            f"{registry_file.as_posix()}:{failed_msgs}"
        )
        return False

    return True


T = TypeVar("T")


@overload
def _filter_by_registries(
    tup: tuple[T, T],
    *,
    home_registry: Literal[True] = ...,
    package_registry: Literal[False] = ...,
    dcopy: bool = ...,
) -> T: ...


@overload
def _filter_by_registries(
    tup: tuple[T, T],
    *,
    home_registry: Literal[False] = ...,
    package_registry: Literal[True] = ...,
    dcopy: bool = ...,
) -> T: ...


@overload
def _filter_by_registries(
    tup: tuple[T, T],
    *,
    home_registry: Literal[True] = ...,
    package_registry: Literal[True] = ...,
    dcopy: bool = ...,
) -> tuple[T, T]: ...


@typecheck
def _filter_by_registries(
    tup: tuple[T, T],
    *,
    home_registry: bool,
    package_registry: bool,
    dcopy: bool = False,
) -> tuple[T, T] | T:
    """
    Filter a tuple by the requested registries.

    :param tup: Tuple to filter
    :type tup: tuple[T, T]
    :param home_registry: Whether to include the home registry
    :type home_registry: bool
    :param package_registry: Whether to include the package registry
    :type package_registry: bool
    :param dcopy: Whether to deepcopy the entries
    :type dcopy: bool
    :return: Filtered tuple or single entry, depending on the input
    :rtype: tuple[T, T] | T
    """
    filtered = []
    if home_registry:
        filtered.append(deepcopy(tup[0]) if dcopy else tup[0])
    if package_registry:
        filtered.append(deepcopy(tup[1]) if dcopy else tup[1])
    return filtered[0] if len(filtered) == 1 else tuple(filtered)



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import filecmp
import shutil
import subprocess
from pathlib import Path

from gurk.lib.context import get_logger
from gurk.lib.context.registry import (
    get_plugin_directories,
    get_plugin_registration,
    is_plugin_registered,
    update_registry,
)
from gurk.lib.core.plugins.virtual_environments import (
    create_venv,
    remove_venv,
    venv_exists,
)
from gurk.lib.shared.configs import load_toml, load_yaml
from gurk.lib.shared.plugins import (
    PluginManifest,
    PluginSource,
    PluginSpecification,
    PluginSpecificationEnum,
)
from gurk.lib.shared.remotes import (
    GitQuery,
    determine_ref,
    edit_url,
    extract_url,
    get_latest_version,
    git_clone,
    is_git_repo,
    parse_git_query,
)
from gurk.lib.utils import (
    GIT_QUERY_VERSIONING_FIELDS,
    GURK_MANIFEST_FILENAME,
    GURK_METADATA_FILENAME,
    Comparison,
    PathLike,
    check_version,
    compare_versions,
    generate_random_path,
    typecheck,
)

from .check import check_local_plugin
from .getters import get_plugin_data, get_relevant_plugin_files
from .versioning import (
    get_local_plugin_version,
    get_plugin_commit,
    get_plugin_version,
)


@typecheck
def is_plugin_installed(
    plugin_spec: PluginSpecification, *, require_venv: bool = True
) -> bool:
    """
    Check if a plugin is validly installed, optionally requiring that its venv exists.

    :param plugin_spec: Name, PathLike, or GitQuery of the plugin. If a GitQuery is provided, the plugin is considered installed if a plugin with the same remote URL and version/commit (if specified) is installed.
    :type plugin_spec: PluginSpecification
    :param require_venv: Whether to check if the plugin's virtual environment exists
    :type require_venv: bool
    :return: True if the plugin is installed (and its venv exists if require_venv is True), False otherwise
    :rtype: bool
    """
    # Get logger
    logger = get_logger()

    # Check that the plugin is validly installed
    try:
        plugin_data = get_plugin_data(plugin_spec)
    except ModuleNotFoundError as e:
        if is_plugin_registered(
            plugin_spec, home_registry=True, package_registry=True
        ):
            logger.debug(
                f"Plugin '{plugin_spec}' is installed but invalid ({e}) - please fix or remove it"
            )
        else:
            logger.debug(f"Plugin '{plugin_spec}' is not installed.")
        return False

    # Check that the specified version/commit is installed
    parsed = parse_git_query(str(plugin_spec))
    if any(parsed[f] for f in GIT_QUERY_VERSIONING_FIELDS) and is_git_repo(
        plugin_spec
    ):
        installed_commit = get_plugin_commit(plugin_spec)
        specified_commit = determine_ref(plugin_spec, to_commit=True)
        if installed_commit != specified_commit:
            msg = (
                f"Plugin '{plugin_spec}' is installed at a "
                "different version than specified:\n- specified: "
            )
            possibly_specified_commit = determine_ref(
                plugin_spec, to_commit=False
            )
            if specified_commit == possibly_specified_commit:
                msg += possibly_specified_commit
            else:
                specified_ref = determine_ref(plugin_spec, to_commit=False)
                msg += f"{specified_ref} -> {specified_commit}"
            logger.debug(f"{msg}\n- installed: {installed_commit}")
            return False

    # Check that the plugin venv exists
    if require_venv and not venv_exists(plugin_data["metadata"]["name"]):
        logger.debug(
            f"Plugin '{plugin_spec}' is installed but its venv is missing."
        )
        return False

    return True


@typecheck
def create_plugin_venv(plugin_name: str) -> bool:
    """
    Create a virtual environment for a plugin based on its dependencies in pyproject.toml.

    :param plugin_name: Name of the plugin
    :type plugin_name: str
    :return: True if the virtual environment was created successfully, False otherwise
    :rtype: bool
    """
    # Get logger
    logger = get_logger()

    # Check if plugin exists locally
    if not is_plugin_installed(plugin_name, require_venv=False):
        logger.error(
            f"Plugin '{plugin_name}' is not installed - Cannot create virtual environment."
        )
        return False

    # Get plugin dependencies
    dependencies = get_plugin_data(plugin_name)["metadata"]["dependencies"]

    # Create plugin venv
    return create_venv(plugin_name, dependencies)


@typecheck
def _install_local_plugin(plugin_path: PathLike) -> bool:
    """
    Import a plugin from a local directory.

    :param plugin_path: Path to the local plugin directory
    :type plugin_path: PathLike
    :return: True if the plugin was imported successfully, False otherwise
    :rtype: bool
    """
    # Get logger
    logger = get_logger()

    # Get plugin manifest
    plugin_path = Path(plugin_path)
    manifest_file = plugin_path / GURK_MANIFEST_FILENAME
    if not manifest_file.is_file():
        logger.error(
            f"Plugin at '{plugin_path}' has no '{GURK_MANIFEST_FILENAME}' file."
        )
        return False

    # Load manifest data
    manifest_data: PluginManifest = load_yaml(manifest_file)
    if not manifest_data:
        raw_manifest = manifest_file.read_text(
            encoding="utf-8", errors="replace"
        )
        logger.error(
            f"Plugin at '{plugin_path}' has empty/invalid YAML in '{GURK_MANIFEST_FILENAME}': "
            f"Preview (max 100 chars):\n{raw_manifest[:100]}"
        )
        return False

    # Get plugin metadata
    metadata = load_toml(plugin_path / GURK_METADATA_FILENAME)
    if not metadata:
        logger.error(
            f"Plugin at '{plugin_path}' has an invalid or missing '{GURK_METADATA_FILENAME}' file",
        )
        return False

    # Extract plugin name from metadata
    try:
        plugin_name: str = metadata["project"]["name"]
    except KeyError as e:
        logger.error(
            f"Plugin at '{plugin_path}' has an invalid '{GURK_METADATA_FILENAME}' file: missing key {e}",
        )
        return False

    # Check if plugin with same name already exists
    if is_plugin_registered(
        plugin_name,
        home_registry=True,
        package_registry=True,
        require_local=True,
    ):
        logger.error(
            f"Plugin with name '{plugin_name}' already exists. Please "
            f"remove it via 'gurk remove {plugin_name}' first."
        )
        return False

    # Pull imported plugins recursively
    for imp in manifest_data.get("imports", []):
        if not is_plugin_installed(extract_url(imp), require_venv=False):
            if not install_plugin(imp):
                logger.error(
                    f"Failed to pull imported plugin '{imp}' for plugin '{plugin_path}'",
                )
                return False
        else:
            if is_plugin_installed(imp, require_venv=False):
                logger.debug(
                    f"Imported plugin '{imp}' for plugin '{plugin_path}' is already installed. Skipping pull."
                )
            else:
                logger.warning(
                    f"Imported plugin '{imp}' for plugin '{plugin_path}' is already installed "
                    f"but with a different version. Please explicitly pull the desired version "
                    f"via 'gurk pull {imp}' to ensure the correct version is installed."
                )

    # Check validity of local plugin
    if not check_local_plugin(plugin_path, verbose=True):
        logger.error(
            f"Plugin at '{plugin_path}' is not a valid gurk plugin.",
        )
        return False

    # Add plugin registry entry
    if not update_registry(plugin_name, {"remote": None}, infer_local=True):
        logger.error(
            f"Failed to add plugin '{plugin_name}' to registry after pulling local plugin from '{plugin_path}'",
        )
        return False
    registration = get_plugin_registration(
        plugin_name,
        home_registry=True,
        package_registry=True,
        require_local=False,
    )
    registration_entry = next(iter(registration.values()))
    dest = registration_entry.get("local")
    if dest is None:
        logger.error(
            f"Registry entry for plugin '{plugin_name}' has no local path, "
            f"although it should have been inferred: {registration_entry}",
        )
        return False

    # Add plugin folder
    if dest.exists():
        logger.warning(
            f"Destination path '{dest}' for plugin '{plugin_name}' already exists. Overwriting it..."
        )
        shutil.rmtree(dest)
    relevant_files = get_relevant_plugin_files(plugin_path)
    for rfile in relevant_files:
        dest_file = dest / rfile
        dest_file.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(plugin_path / rfile, dest_file)

    # Install plugin dependencies in the plugin venv
    if not create_plugin_venv(plugin_name):
        logger.error(
            f"Failed to create virtual environment for plugin '{plugin_name}'",
        )
        return False

    # Verify installation
    if not is_plugin_installed(plugin_name, require_venv=True):
        logger.error(
            f"Plugin '{plugin_name}' installation verification "
            f"failed after pulling local plugin from '{plugin_path}'",
        )
        return False

    return True


@typecheck
def _install_remote_plugin(plugin: GitQuery) -> bool:
    """
    Import a plugin from a remote Git repository.

    :param plugin: GitQuery of the plugin to import
    :type plugin: GitQuery
    :return: True if the plugin was imported successfully, False otherwise
    :rtype: bool
    """

    def error(message: str, _temp_path: Path | None = None):
        """
        Log an error message and clean up temporary plugin path if provided.

        :param message: Error message to log
        :type message: str
        :param _temp_path: Temporary plugin path to clean up
        :type _temp_path: Path | None
        """
        get_logger().error(message + ". Skipping...")
        if _temp_path is not None and _temp_path.exists():
            if _temp_path.is_dir():
                shutil.rmtree(_temp_path)
            else:
                _temp_path.unlink()

    # Check that the repo exists
    if not is_git_repo(plugin):
        error(
            f"Remote plugin source '{plugin}' does not exist or is not a valid git repository"
        )
        return False

    # Check that version/commit specified exists
    commit = determine_ref(plugin, to_commit=True)
    if commit is None:
        error(
            f"Specified version/commit in remote plugin source '{plugin}' does not exist"
        )
        return False

    # Import metadata to random file
    temp_metadata = generate_random_path(suffix=".toml", create=False)
    try:
        git_clone(
            edit_url(plugin, path=GURK_METADATA_FILENAME),
            temp_metadata,
        )
    except subprocess.CalledProcessError as e:
        error(
            f"Failed to clone '{GURK_METADATA_FILENAME}' from "
            f"remote plugin repository '{plugin}': {e}",
            temp_metadata,
        )
        return False
    except ValueError as e:
        error(str(e), temp_metadata)
        return False

    # Get relevant metadata
    try:
        metadata = load_toml(temp_metadata)
        plugin_name: str = metadata["project"]["name"]
        plugin_version: str = metadata["project"]["version"]
        if not check_version(plugin_version):
            raise ValueError(f"Invalid version string: {plugin_version}")
    except KeyError as e:
        error(
            f"Plugin '{plugin}' has an invalid '{GURK_METADATA_FILENAME}' file: invalid key {e}",
            temp_metadata,
        )
        return False

    # Check if plugin already exists
    if is_plugin_installed(plugin_name):
        error(
            f"Plugin with remote '{plugin}' is already installed. "
            f"Please remove it via 'gurk remove {plugin_name}' "
            f"first or update it via 'gurk update {plugin_name}'"
        )
        return False

    # See if plugin source is valid
    if not is_git_repo(plugin):
        error(
            f"Remote plugin source '{plugin}' does not exist or is not a valid git repository"
        )
        return False

    # Clone only relevant files to temporary directory
    relevant_files = get_relevant_plugin_files(plugin, relative=False)
    if relevant_files is None:
        error(
            f"Failed to pull relevant files for plugin '{plugin}' from its remote repository"
        )
        return False
    temp_plugin_path = [
        f for f in relevant_files if f.name == GURK_MANIFEST_FILENAME
    ][0].parent

    # Pull local plugin from temporary directory
    if not _install_local_plugin(temp_plugin_path):
        error(
            f"Failed to import plugin from remote repository '{plugin}'",
            temp_plugin_path,
        )
        return False

    # Upate plugin registry entry to include remote
    if not update_registry(
        plugin_name,
        {
            "remote": edit_url(
                extract_url(plugin),
                commit=commit,
            )
        },
    ):
        error(
            f"Failed to update registry entry for plugin '{plugin_name}' with remote information after pulling from '{plugin}'",
        )
        return False

    # Clean up temporary plugin path
    shutil.rmtree(temp_plugin_path)

    return True


@typecheck
def remove_plugin(plugin: PluginSpecification, verbose: bool = False) -> None:
    """
    Remove a locally installed plugin.

    :param plugin: Name, PathLike, or GitQuery of the plugin to remove
    :type plugin: PluginSpecification
    :param verbose: Whether to print info messages
    :type verbose: bool
    :raises ModuleNotFoundError: If no such local plugin is found
    """
    # Get logger
    logger = get_logger()
    remove_msg = []

    # Get plugin data
    plugin_registration = get_plugin_registration(
        plugin, home_registry=True, package_registry=True, require_local=False
    )
    if not plugin_registration:
        raise ModuleNotFoundError(
            f"No installed plugin called '{plugin}' found"
        )
    plugin_name, plugin_entry = next(iter(plugin_registration.items()))

    # Remove plugin registry entry
    if update_registry(plugin_name, None):
        remove_msg.append("registry entry")

    # Remove plugin folder
    if plugin_entry["local"] and plugin_entry["local"].is_dir():
        shutil.rmtree(plugin_entry["local"])
        remove_msg.append("plugin files")

    # Remove plugin venv
    if remove_venv(plugin_name):
        remove_msg.append("virtual environment")

    if verbose:
        if remove_msg:
            logger.success(
                f"Successfully removed {' and '.join(remove_msg)} for plugin '{plugin_name}'"
            )
        else:
            logger.info(
                f"Nothing to remove for (package) plugin '{plugin_name}'"
            )


@typecheck
def install_plugin(
    plugin_source: PluginSource, reinstall: bool = False
) -> bool:
    """
    Install a plugin from a specified source, which can be either a local path or a Git URL.

    :param plugin_source: Path or Git URL specifying the plugin source
    :type plugin_source: PluginSource
    :param reinstall: Whether to reinstall the plugin if it is already installed (but does not match the specified source or version).
    :type reinstall: bool
    :return: True if the plugin was (re)installed successfully, False otherwise
    :rtype: bool
    """
    # Get logger
    logger = get_logger()

    def unexpected_source_type_error():
        logger.error(
            f"Unexpected: Unknown source type '{source_type}' for source '{plugin_source}'."
        )

    # Check source type
    if Path(plugin_source).is_dir():
        source_type = PluginSpecificationEnum.LOCAL_PATH
    elif is_git_repo(plugin_source):
        source_type = PluginSpecificationEnum.GIT_REMOTE
    else:
        logger.error(
            f"Invalid plugin source '{plugin_source}': Must be either a local "
            "path or a Git Query whose URL points to an existing repository."
        )
        return False

    # Check source and get plugin spec
    if source_type == PluginSpecificationEnum.LOCAL_PATH:
        plugin_path = Path(plugin_source)
        # Check that the local path is not under either plugin directory
        if any(
            plugin_path.is_relative_to(d) for d in get_plugin_directories()
        ):
            logger.error(
                f"Specified local path '{plugin_path}' is invalid: "
                "Local paths cannot be under plugin directories."
            )
            return False

        # Get the plugin name
        try:
            plugin_spec = load_toml(plugin_path / GURK_METADATA_FILENAME)[
                "project"
            ]["name"]
        except Exception as e:
            logger.error(
                f"Failed to load plugin name from '{plugin_path}': {str(e)}"
            )
            return False

        # Check that there is no plugin with the same name registered with a remote
        if is_plugin_registered(
            plugin_spec,
            home_registry=True,
            package_registry=True,
            require_local=False,
        ):
            registration = get_plugin_registration(
                plugin_spec,
                home_registry=True,
                package_registry=True,
                require_local=False,
            )
            entry = next(iter(registration.values()))
            if entry.get("remote") is not None:
                logger.error(
                    "A plugin with the same name as the specified local plugin already "
                    f"exists with a remote URL registered ({entry['remote']}), so the "
                    "two plugins are considered different. Please change the specification "
                    f"or remove the existing plugin first using 'gurk remove {plugin_spec}'."
                )
                return False
    elif source_type == PluginSpecificationEnum.GIT_REMOTE:
        # Check that max one of version/commit is specified in the remote URL (if any)
        parsed = parse_git_query(plugin_source)
        if (
            len(
                [
                    f
                    for f in GIT_QUERY_VERSIONING_FIELDS
                    if parsed[f] is not None
                ]
            )
            > 1
        ):
            logger.error(
                f"Invalid Git remote source '{plugin_source}':"
                " Cannot specify more than one of "
                f"{', '.join(GIT_QUERY_VERSIONING_FIELDS)} in the Git remote URL."
            )
            return False
        plugin_spec = plugin_source
    else:
        unexpected_source_type_error()
        return False

    # Utility function to install plugin based on source type
    def _install_plugin() -> bool:
        """
        Install the plugin based on its source type.

        :return: True if the plugin was installed successfully, False otherwise
        :rtype: bool
        """
        logger.info(f"Installing plugin from '{plugin_source}'...")
        if source_type == PluginSpecificationEnum.LOCAL_PATH:
            success = _install_local_plugin(plugin_source)
        elif source_type == PluginSpecificationEnum.GIT_REMOTE:
            success = _install_remote_plugin(plugin_source)
        else:
            unexpected_source_type_error()
            return False

        if success:
            logger.success(
                f"Successfully installed plugin from '{plugin_source}'"
            )
        else:
            logger.error(f"Failed to install plugin from '{plugin_source}'")
        return success

    # Act depending on whether the plugin is already installed
    if not is_plugin_installed(extract_url(plugin_spec), require_venv=False):
        # Handle existing venv
        if venv_exists(plugin_spec):
            if not reinstall:
                logger.error(
                    f"Plugin '{plugin_spec}' is not currently installed but "
                    "has an existing virtual environment. Pass 'reinstall=True' "
                    "to remove the existing venv and proceed with installation."
                )
                return False
            else:
                logger.debug(
                    f"Removing existing virtual environment for plugin '{plugin_spec}' to proceed with installation..."
                )
                if not remove_venv(plugin_spec):
                    logger.error(
                        f"Failed to remove existing virtual environment for plugin '{plugin_spec}'."
                    )
                    return False

        # Install plugin
        logger.info(f"Plugin '{plugin_spec}' is not currently installed.")
        if not _install_plugin():
            return False
    else:
        # Get plugin data
        plugin_data = get_plugin_data(plugin_spec)

        # Determine if the plugin needs to be reinstalled
        reinstall_required = False
        if source_type == PluginSpecificationEnum.LOCAL_PATH:
            if reinstall:
                reinstall_required = True
            else:
                source_version = get_local_plugin_version(plugin_source)
                installed_version = get_plugin_version(
                    plugin_spec, require_local=True
                )
                if (
                    compare_versions(source_version, installed_version)
                    == Comparison.EQUAL
                ):
                    logger.info(
                        f"Plugin '{plugin_spec}' is already installed and "
                        "matches the specified local source version "
                        f"({source_version}). Skipping installation..."
                    )
                    return True
                else:
                    logger.error(
                        f"Plugin '{plugin_spec}' is already installed (v: {installed_version})"
                        ", but does not match the specified local source version "
                        f"(v: {source_version}). Please explictly specify to reinstall it."
                    )
                    return False
        elif source_type == PluginSpecificationEnum.GIT_REMOTE:
            # Check version/commit against existing one if specified
            if not plugin_data["registration"]["remote"]:
                logger.error(
                    f"Plugin '{plugin_spec}' is installed but does not have a "
                    "remote URL registered, so the two plugins are considered "
                    "different. Please change the specification or remove the "
                    f"existing plugin first using 'gurk remove {plugin_spec}'."
                )
                return False
            if not is_plugin_installed(plugin_spec, require_venv=False):
                if reinstall:
                    reinstall_required = True
                else:
                    logger.error(
                        f"Plugin '{plugin_spec}' is installed at a different "
                        "version than specified (see full log). Use '--replace' "
                        "to reinstall and use the specified version."
                    )
                    return False
        else:
            unexpected_source_type_error()
            return False

        # Reinstall plugin if needed
        if reinstall_required:
            logger.info(
                f"Reinstalling plugin '{plugin_spec}' to match specified commit."
            )
            # Remove existing plugin
            try:
                remove_plugin(plugin_spec, verbose=True)
            except ModuleNotFoundError as e:
                logger.error(str(e))
                return False
            # Install plugin
            if not _install_plugin():
                return False
        else:
            # Install the plugin venv if it doesn't exist
            plugin_name = plugin_data["metadata"]["name"]
            if not venv_exists(plugin_name):
                logger.info(
                    f"Plugin '{plugin_spec}' is already installed but has no virtual environment. Creating venv..."
                )
                if not create_plugin_venv(plugin_name):
                    logger.error(
                        f"Failed to create virtual environment for plugin '{plugin_spec}'."
                    )
                    return False

    return True


@typecheck
def upgrade_plugin(
    plugin_spec: PluginSpecification, require_local: bool = True
) -> bool:
    """
    Upgrade an installed plugin to the latest version available at its registered remote URL.

    :param plugin_spec: Name, PathLike, or GitQuery of the plugin to upgrade. If a GitQuery is provided, it is only used to identify the plugin to upgrade.
    :type plugin_spec: PluginSpecification
    :param require_local: Whether to only consider plugins with a local installation when looking for the plugin to upgrade. When upgrading a remote-only plugin, it is not installed here, but its registry entry is updated to point to the new version.
    :type require_local: bool
    :return: True if the plugin was upgraded successfully, False otherwise
    :rtype: bool
    """
    # Get logger
    logger = get_logger()

    # Check that the plugin is registered
    if not is_plugin_registered(
        plugin_spec,
        home_registry=True,
        package_registry=True,
        require_local=require_local,
    ):
        logger.debug(f"Plugin '{plugin_spec}' is not registered. Skipping...")
        return False

    # Check that the plugin has a registered remote URL
    plugin_remote = next(
        iter(
            get_plugin_registration(
                plugin_spec,
                home_registry=True,
                package_registry=True,
                require_local=require_local,
            ).values()
        )
    )["remote"]
    if plugin_remote is None:
        logger.error(f"Plugin '{plugin_spec}' is local-only. Skipping...")
        return False

    # Get the latest version available at the registered remote URL
    latest_version = get_latest_version(plugin_remote)
    if latest_version is None:
        logger.error(
            f"Failed to get the latest version for plugin '{plugin_spec}' from its registered remote URL. Skipping..."
        )
        return False

    # Get the current registered version (if any)
    current_version = get_plugin_version(
        plugin_spec, require_local=require_local
    )
    if current_version is None:
        logger.error(
            f"Failed to get the registered version for plugin '{plugin_spec}'. Skipping..."
        )
        return False

    # # Check if the registered version is already the latest version
    version_comparison = compare_versions(latest_version, current_version)
    if version_comparison is None:
        logger.error(
            f"Failed to compare versions for plugin '{plugin_spec}'. Skipping..."
        )
        return False
    elif version_comparison == Comparison.EQUAL:
        logger.info(
            f"Plugin '{plugin_spec}' is already at the latest version ({current_version})."
        )
        return True
    elif version_comparison == Comparison.SMALLER:
        logger.warning(
            f"Unexpected: Plugin '{plugin_spec}' version ({current_version}) "
            f"is newer than latest ({latest_version}). Skipping..."
        )
        return False
    else:
        logger.info(
            f"Upgrading plugin '{plugin_spec}' from {current_version} to {latest_version}..."
        )
        # Get the current relevant files
        current_relevant_files = get_relevant_plugin_files(plugin_spec)
        if current_relevant_files is None:
            logger.error(
                f"Failed to get relevant files for plugin '{plugin_spec}'."
            )
            return False
        current_remote = edit_url(
            extract_url(plugin_remote), version=current_version
        )

        # Get the latest relevant files
        latest_relevant_files = get_relevant_plugin_files(plugin_remote)
        if latest_relevant_files is None:
            logger.error(
                f"Failed to pull relevant files for plugin '{plugin_spec}'."
            )
            return False
        latest_remote = edit_url(
            extract_url(plugin_remote), version=latest_version
        )

        # Check if the relevant files have changed
        if current_relevant_files == latest_relevant_files:
            abs_current_relevant_files = get_relevant_plugin_files(
                current_remote, relative=False
            )
            abs_latest_relevant_files = get_relevant_plugin_files(
                latest_remote, relative=False
            )
            if len(abs_current_relevant_files) != len(
                abs_latest_relevant_files
            ):
                files_differ = True
            else:
                files_differ = False
                for cfile, lfile in zip(
                    sorted(current_relevant_files),
                    sorted(latest_relevant_files),
                ):
                    if (
                        cfile.name == GURK_METADATA_FILENAME
                        and lfile.name == GURK_METADATA_FILENAME
                    ):
                        # Skip comparing pyproject.toml, since we already know it has a version change
                        #   Dependency changes are reflected in other changed files
                        continue
                    elif not filecmp.cmp(
                        cfile,
                        lfile,
                        shallow=False,
                    ):
                        files_differ = True
                        break

            if not files_differ:
                logger.info(
                    f"Plugin '{plugin_spec}' files unchanged. Skipping..."
                )
                return True

    # Handle upgrading to the latest version based on if the plugin is installed
    if not is_plugin_installed(extract_url(plugin_spec), require_venv=False):
        # Update registry entry to point to the latest version at the registered remote URL
        if not update_registry(
            plugin_spec,
            {"remote": latest_remote},
        ):
            logger.error(
                f"Failed to update registry entry for plugin '{plugin_spec}' to version {latest_version}."
            )
            return False
    else:
        # Reinstall plugin at the latest version
        if not install_plugin(latest_remote, reinstall=True):
            logger.error(
                f"Failed to upgrade plugin '{plugin_spec}' to version {latest_version}."
            )
            return False

    return True



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa: F401
from .data import PluginData
from .manifest import PluginManifest, PluginOptions, ResolvedPluginManifest
from .metadata import (
    PluginMetadata,
    PluginMetadataDependencies,
    PluginMetadataProject,
    ResolvedPluginMetadata,
)
from .registry import (
    PluginRegistry,
    PluginRegistryEntry,
    ResolvedPluginRegistry,
    ResolvedPluginRegistryEntry,
    ResolvedZippedRegistry,
    ZippedRegistry,
)
from .specification import (
    PluginSource,
    PluginSpecification,
    PluginSpecificationEnum,
)



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shutil
import subprocess
from pathlib import Path
from typing import Iterator

from gurk.lib.context.logger import get_logger
from gurk.lib.context.registry import (
    get_available_plugin_names,
    get_plugin_registration,
)
from gurk.lib.shared.configs import load_toml, load_yaml
from gurk.lib.shared.dicts import fill_typed_dict, filter_typed_dict
from gurk.lib.shared.plugins import (
    PluginData,
    PluginManifest,
    PluginMetadata,
    PluginSource,
    PluginSpecification,
    ResolvedPluginManifest,
    ResolvedPluginMetadata,
)
from gurk.lib.shared.remotes import GitQuery, edit_url, git_clone, is_git_repo
from gurk.lib.shared.tasks import (
    ResolvedDefaultTaskDictCollection,
    TaskDictCollection,
)
from gurk.lib.utils import (
    GURK_MANIFEST_FILENAME,
    GURK_METADATA_FILENAME,
    PathLike,
    full_isinstance,
    generate_random_path,
    typecheck,
)

from .check import check_local_plugin


@typecheck
def get_raw_plugin_manifest(
    plugin: PluginSpecification,
) -> PluginManifest | None:
    """
    Get the raw manifest of a plugin if it exists locally.

    :param plugin: Name, utils.PathLike, or GitQuery of the plugin
    :type plugin: PluginSpecification
    :return: Plugin manifest if the plugin exists locally, None otherwise
    :rtype: PluginManifest | None
    """
    plugin_registration = get_plugin_registration(
        plugin, home_registry=True, package_registry=True
    )
    if not plugin_registration:
        return None
    plugin_registration_entry = next(iter(plugin_registration.values()))

    if not check_local_plugin(plugin_registration_entry["local"]):
        return None

    raw_plugin_yaml = load_yaml(
        plugin_registration_entry["local"] / GURK_MANIFEST_FILENAME
    )
    if raw_plugin_yaml is None:
        return None

    return raw_plugin_yaml


@typecheck
def get_resolved_plugin_manifest(
    plugin: PluginSpecification,
) -> ResolvedPluginManifest | None:
    """
    Get the manifest of a local plugin with
    - all paths resolved and converted to "Path" objects
    - missing properties filled with default values

    :param plugin: Name, utils.PathLike, or GitQuery of the plugin
    :type plugin: PluginSpecification
    :return: Plugin configuration with resolved paths and filled properties if the plugin exists locally, None otherwise
    :rtype: ResolvedPluginManifest | None
    """
    plugin_manifest = get_raw_plugin_manifest(plugin)
    if not plugin_manifest:
        return None

    # Fill missing properties
    plugin_manifest = fill_typed_dict(plugin_manifest, ResolvedPluginManifest)

    # Expand task paths
    plugin_registration = get_plugin_registration(
        plugin, home_registry=True, package_registry=True
    )
    if not plugin_registration:
        return None
    plugin_registration_entry = next(iter(plugin_registration.values()))
    for _, task in plugin_manifest["tasks"].items():
        # Expand script path
        task["script"] = plugin_registration_entry["local"] / task["script"]

        # Expand config_file path (if applicable)
        if task["config_file"] is not None:
            task["config_file"] = (
                plugin_registration_entry["local"] / task["config_file"]
            )

    # Expand option task paths
    for _, option in plugin_manifest["options"].items():
        for _, task in option.items():
            # Expand config_file path (if applicable)
            if task["config_file"] is not None:
                task["config_file"] = (
                    plugin_registration_entry["local"] / task["config_file"]
                )

    return plugin_manifest


@typecheck
def _get_plugin_metadata(
    plugin: PluginSpecification,
) -> ResolvedPluginMetadata | None:
    """
    Get the pyproject.toml metadata of a local plugin.

    :param plugin: Name, utils.PathLike, or GitQuery of the plugin
    :type plugin: PluginSpecification
    :return: Plugin metadata if the plugin exists locally, None otherwise
    :rtype: ResolvedPluginMetadata | None
    """
    plugin_registration = get_plugin_registration(
        plugin, home_registry=True, package_registry=True
    )
    if not plugin_registration:
        return None
    plugin_registration_entry = next(iter(plugin_registration.values()))

    if not check_local_plugin(plugin_registration_entry["local"]):
        return None

    toml_data = load_toml(
        plugin_registration_entry["local"] / GURK_METADATA_FILENAME
    )
    if not toml_data:
        return None

    metadata = filter_typed_dict(toml_data, PluginMetadata)["project"]
    dependencies = metadata.pop("optional-dependencies", {})
    metadata["dependencies"] = dependencies.get("gurk", [])

    return metadata


@typecheck
def get_plugin_data(
    plugin: PluginSpecification,
) -> PluginData:
    """
    Get the registry entry, manifest and pyproject.toml metadata of a local plugin.

    :param plugin: Name, utils.PathLike, or GitQuery of the plugin
    :type plugin: PluginSpecification
    :return: Plugin data containing registry entry, manifest and metadata
    :rtype: PluginData
    :raises ModuleNotFoundError: If no valid plugin was found
    """

    def error_msg(message: str) -> str:
        return f"ERROR loading plugin data for {plugin}: {message}"

    plugin_registration = get_plugin_registration(
        plugin, home_registry=True, package_registry=True
    )
    if not plugin_registration:
        raise ModuleNotFoundError(
            error_msg("Could not load a (valid) plugin entry")
        )

    plugin_manifest = get_resolved_plugin_manifest(plugin)
    if not plugin_manifest:
        raise ModuleNotFoundError(
            error_msg(
                f"Could not load a (valid) {GURK_MANIFEST_FILENAME} file"
            )
        )

    plugin_metadata = _get_plugin_metadata(plugin)
    if not plugin_metadata:
        raise ModuleNotFoundError(
            error_msg(
                "UNEXPECTED: Could not load plugin metadata from a (valid) pyproject.toml file"
            )
        )

    return PluginData(
        registration=next(iter(plugin_registration.values())),
        manifest=plugin_manifest,
        metadata=plugin_metadata,
    )


def get_available_plugin_tasks() -> ResolvedDefaultTaskDictCollection:
    """
    Get the combined task definitions of all local plugins.

    :return: Dictionary of tasks from all local plugins
    :rtype: ResolvedDefaultTaskDictCollection
    """
    combined_tasks: TaskDictCollection = {}
    for plugin in get_available_plugin_names():
        plugin_manifest = get_resolved_plugin_manifest(plugin)
        if plugin_manifest:
            combined_tasks.update(plugin_manifest["tasks"])

    return combined_tasks


def iter_scripts() -> Iterator[Path]:
    """
    Yields all script files of installed plugins.

    :return: Iterator of Paths to script files
    :rtype: Iterator[Path]
    """
    all_tasks = get_available_plugin_tasks()
    for task in all_tasks.values():
        yield task["script"]


def iter_configs() -> Iterator[Path]:
    """
    Yields all config files of installed plugins.

    :return: Iterator of Paths to config files
    :rtype: Iterator[Path]
    """
    all_tasks = get_available_plugin_tasks()
    for task in all_tasks.values():
        if task.get("config_file"):
            yield task["config_file"]


@typecheck
def _get_relevant_files_local(
    plugin_path: PathLike, *, relative: bool = True
) -> set[Path] | None:
    """
    Get the set of relevant plugin files (script, config_file, manifest) for a plugin given its local path.

    :param plugin_path: Path to a local directory containing a plugin manifest
    :type plugin_path: PathLike
    :param relative: Whether to return paths relative to the plugin path
    :type relative: bool
    :return: Set of Paths to relevant plugin files
    :rtype: set[Path] | None
    """

    def rel_path(path_str: str) -> Path:
        return (
            Path(path_str)
            if relative
            else (Path(plugin_path) / path_str).expanduser().resolve()
        )

    plugin_manifest = Path(plugin_path) / GURK_MANIFEST_FILENAME
    if not plugin_manifest.exists():
        return None

    raw_plugin_yaml = load_yaml(plugin_manifest)
    if not raw_plugin_yaml or not full_isinstance(
        raw_plugin_yaml, PluginManifest
    ):
        return None

    # Defined tasks
    relevant_files = {
        rel_path(GURK_MANIFEST_FILENAME),
        rel_path(GURK_METADATA_FILENAME),
    }
    for task in raw_plugin_yaml.get("tasks", {}).values():
        # 'script'
        relevant_files.add(rel_path(task["script"]))
        # 'config_file' (if applicable)
        if task.get("config_file"):
            relevant_files.add(rel_path(task["config_file"]))

    # Option tasks
    for option in raw_plugin_yaml.get("options", {}).values():
        for task in option.values():
            # 'config_file' (if applicable)
            if task.get("config_file"):
                relevant_files.add(rel_path(task["config_file"]))

    return relevant_files


@typecheck
def _get_relevant_files_remote(
    plugin_remote: str | GitQuery, *, relative: bool = True
) -> set[Path] | None:
    """
    Get the set of relevant plugin files (script, config_file, manifest) for a plugin given its remote source.

    :param plugin_remote: Remote source of the plugin
    :type plugin_remote: str | GitQuery
    :param relative: Whether to return paths relative to the plugin path
    :type relative: bool
    :return: Set of Paths to relevant plugin files
    :rtype: set[Path] | None
    """
    if not is_git_repo(plugin_remote):
        return None

    # Import manifest to random file
    temp_dir = generate_random_path(prefix="gurk_plugin_", create=True)
    temp_manifest = temp_dir / GURK_MANIFEST_FILENAME
    try:
        git_clone(
            edit_url(plugin_remote, path=GURK_MANIFEST_FILENAME),
            temp_manifest,
        )
    except subprocess.CalledProcessError:
        return None

    # Determine relevant files
    relevant_files = _get_relevant_files_local(temp_dir, relative=relative)

    # Clone all relevant files to temporary directory if absolute paths are requested
    def reverse_rel_path(path: Path) -> Path:
        return path if relative else path.relative_to(temp_dir)

    if not relative and relevant_files is not None:
        for file in relevant_files:
            if file.name == GURK_MANIFEST_FILENAME:
                continue  # Manifest is already imported
            relative_file = str(reverse_rel_path(file))
            try:
                git_clone(
                    edit_url(plugin_remote, path=relative_file),
                    temp_dir / relative_file,
                )
            except subprocess.CalledProcessError as e:
                get_logger().error(
                    f"Failed to pull relevant file '{relative_file}' for plugin from its remote repository: {e}"
                )
                return None

    # Cleanup - if only relative paths are asked for, remove the temporary directory
    if relative and temp_dir.exists():
        shutil.rmtree(temp_dir)

    return relevant_files


@typecheck
def get_relevant_plugin_files(
    plugin_source: PluginSource, *, relative: bool = True
) -> set[Path] | None:
    """
    Get the set of relevant plugin files (script, config_file, manifest) for a plugin given its source.

    :param plugin_source: Source of the plugin
    :type plugin_source: PluginSource
    :param relative: Whether to return paths relative to the plugin path. If False, returns absolute paths and clones files from the remote if necessary.
    :type relative: bool
    :return: Set of Paths to relevant plugin files
    :rtype: set[Path] | None
    :raises ValueError: If the plugin source is invalid
    """
    if Path(plugin_source).exists():
        return _get_relevant_files_local(plugin_source, relative=relative)
    elif is_git_repo(str(plugin_source)):
        return _get_relevant_files_remote(
            str(plugin_source), relative=relative
        )
    else:
        raise ValueError(
            f"Invalid plugin source '{plugin_source}' - must be a git remote or a "
            f"local plugin path, not {plugin_source} (type '{type(plugin_source)}')"
        )



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import TypeAlias, TypedDict

from ..remotes import GitQuery


class PluginRegistryEntry(TypedDict):
    # fmt: off
    local:   None | str
    remote:  None | GitQuery
    # fmt: on


PluginRegistry: TypeAlias = dict[str, PluginRegistryEntry]
ZippedRegistry: TypeAlias = tuple[Path, PluginRegistry]


class ResolvedPluginRegistryEntry(TypedDict):
    # fmt: off
    local:   None | Path
    remote:  None | GitQuery
    # fmt: on


ResolvedPluginRegistry: TypeAlias = dict[str, ResolvedPluginRegistryEntry]
ResolvedZippedRegistry: TypeAlias = tuple[Path, ResolvedPluginRegistry]



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Any

import tomli_w
import tomllib

from gurk.lib.utils import PathLike, typecheck


@typecheck
def load_toml(
    toml_file_or_str: PathLike, *, from_str: bool = False
) -> dict[str, Any] | None:
    """
    Load TOML content.

    :param toml_file_or_str: Path to the TOML file to load or a string containing TOML content (if 'from_str' is True)
    :type toml_file_or_str: PathLike
    :param from_str: Whether 'toml_file_or_str' is a string containing TOML content (True) or a file path (False, default)
    :type from_str: bool
    :return: Content of the TOML file, or None if loading fails
    :rtype: dict[str, Any] | None
    """
    if from_str:
        try:
            return tomllib.loads(toml_file_or_str) or {}
        except tomllib.TOMLDecodeError:
            return None
    else:
        if not Path(toml_file_or_str).is_file():
            return None
        with open(toml_file_or_str, "rb") as f:
            try:
                return tomllib.load(f) or {}
            except tomllib.TOMLDecodeError:
                return None


@typecheck
def dump_toml(content: dict[str, Any], toml_file: PathLike) -> None:
    """
    Dump content to a TOML file.

    :param content: Content to dump
    :type content: dict[str, Any]
    :param toml_file: Path to the TOML file to dump to
    :type toml_file: PathLike
    """
    with open(toml_file, "wb") as f:
        tomli_w.dump(content, f)



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
from pathlib import Path
from shutil import copy2, copytree
from typing import Any

import requests

from gurk import (
    LoggerSeverity,
    PatternCollection,
    get_clean_lines,
    git_clone,
    is_git_repo,
    is_url,
    load_yaml,
    log_step,
    logrichprint,
    parse_task_args,
    resolve_package_path,
    revert_sudo_permissions,
)


def configure_pinned_apps(*args: list[str]) -> None:
    """
    Configure pinned applications in the GNOME desktop environment.

    NOTE: Inexistent apps are stored as "pinned" but only actually pin after being installed.
          No error is raised by 'gsettings' if an app does not exist.
    """
    # Parse task args
    task_args = parse_task_args(args)

    # Get apps to pin
    apps = get_clean_lines(task_args.config_file)
    apps_str = "['" + "', '".join(apps) + "']"

    # (STEP) Pinning apps...
    subprocess.run(
        ["gsettings", "set", "org.gnome.shell", "favorite-apps", apps_str]
    )


# TODO: Expand any "[/].../.../..." paths to dicts. Pay attention not to merge with existing dicts.
#       This could also be used together with getent_passwd(os.getenv("SUDO_USER"), "home")
#           to ge the home directory of the user and thus remove the need to use HOME and ROOT.
def configure_filestructure(*args: list[str]) -> None:
    """
    Create a predefined file structure based on a YAML mapping.
    """
    # Parse task args
    task_args = parse_task_args(args)

    def recursive_create_structure(
        base_path: Path, structure: dict[str, Any], overwrite: bool, sudo: bool
    ) -> None:
        for name, content in structure.items():
            dest_path = base_path / name
            if (content is None or isinstance(content, str)) and (
                dest_path.exists() and not overwrite
            ):
                log_step(
                    f"Path {dest_path} already exists. Skipping creation...",
                    warning=True,
                )
                continue

            if content is None:
                if Path(name).suffix:
                    # It's a file
                    dest_path.touch(exist_ok=True)
                else:
                    # It's a directory
                    dest_path.mkdir(exist_ok=True)
            elif isinstance(content, str):
                # Detect symlink
                symlink_match = PatternCollection.PATH.patterns[
                    "symlink"
                ].match(content)

                # Resolve package path (if applicable)
                content = resolve_package_path(
                    content if not symlink_match else symlink_match.group(1)
                )
                if content is None:
                    log_step(
                        f"Package resource in path '{content}' could not be resolved. Skipping...",
                        warning=True,
                    )
                    continue

                # Get content based on type
                if is_git_repo(content):
                    log_step(
                        f"Cloning git repository {content} into {dest_path}..."
                    )
                    try:
                        git_clone(content, dest_path, overwrite)
                    except subprocess.CalledProcessError:
                        log_step(
                            f"Failed to clone git repository {content}. Skipping...",
                            warning=True,
                        )
                        continue
                elif is_url(content):
                    log_step(
                        f"Downloading file from {content} to {dest_path}..."
                    )
                    response = requests.get(
                        content, timeout=60, headers={"Accept-Encoding": "*"}
                    )
                    if response.status_code == 200:
                        dest_path.write_bytes(response.content)
                    else:
                        log_step(
                            f"Failed to download file from {content}. HTTP status code: {response.status_code}",
                            warning=True,
                        )
                else:
                    # Assumed local path (possibly symlinked)
                    content = Path(
                        content
                    ).expanduser()  # TODO: Does this work when this is working in privileged mode?
                    ## Absolute path
                    if content.is_absolute():
                        pass
                    ## Relative path (to base path)
                    elif str(content).startswith("./"):
                        content = (base_path / content).resolve()
                    ## Relative path (to config file)
                    else:
                        content = (
                            task_args.config_file.parent / content
                        ).resolve()

                    if not content.exists():
                        log_step(
                            f"Source '{content}' does not exist. Skipping...",
                            warning=True,
                        )
                        continue

                    if symlink_match:
                        log_step(
                            f"Creating symlink from {content} to {dest_path}..."
                        )
                        dest_path.symlink_to(content)
                    else:
                        log_step(
                            f"Copying from local path {content} to {dest_path}..."
                        )
                        if content.is_file():
                            copy2(content, dest_path)
                        elif content.is_dir():
                            copytree(content, dest_path, dirs_exist_ok=True)
            elif isinstance(content, dict):
                # It's a directory with further contents
                dest_path.mkdir(exist_ok=True)
                recursive_create_structure(dest_path, content, overwrite, sudo)
            else:
                log_step(
                    f"Unsupported entry type '{type(content).__name__}' for {content}. Skipping...",
                    warning=True,
                )

            # Revert to user permissions if under HOME directory
            if not sudo:
                revert_sudo_permissions(dest_path)

    # Check file structure
    config_data = load_yaml(task_args.config_file)
    if config_data is None:
        logrichprint(
            LoggerSeverity.FATAL,
            f"Invalid YAML file provided for file structure configuration: {task_args.config_file}",
        )
        raise ValueError

    # (STEP) Creating file structure...
    if config_data.get("HOME"):
        recursive_create_structure(
            Path.home(),
            config_data["HOME"],
            task_args.force,
            False,
        )
    if config_data.get("ROOT"):
        if not task_args.gurk_configure_root_filestructure:
            logrichprint(
                LoggerSeverity.WARNING,
                "Skipping root (/) file structure configuration, as "
                "'--gurk-configure-root-filestructure' flag is not provided.",
            )
        else:
            recursive_create_structure(
                Path("/"), config_data["ROOT"], False, True
            )



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
from pathlib import Path
from typing import Literal

from gurk.lib.shared.scripts import run_script_function
from gurk.lib.utils import PACKAGE_BASH_HELPERS_PATH, typecheck

from .interface import log_step

PasswdField = Literal[
    "username",
    "uid",
    "gid",
    "gecos",
    "home",
    "shell",
]


@typecheck
def getent_passwd(
    user: str | int,
    field: PasswdField,
) -> str:
    """
    Retrieve specific fields from the system's user database for a given user.

    :param user: The username or user ID to look up
    :type user: str | int
    :param field: The field to retrieve from the user's passwd entry
    :type field: PasswdField
    :return: The requested field from the user's passwd entry
    :rtype: str
    :raises ValueError: If an invalid field is specified or the user is not found
    """
    if field not in PasswdField.__args__:
        raise ValueError(f"Invalid field: {field!r}")

    # Retrieve the passwd entry using getent
    try:
        output = subprocess.check_output(
            ["getent", "passwd", str(user)],
            text=True,
        )
    except subprocess.CalledProcessError:
        raise ValueError(f"User {user!r} not found")

    # Parse the output and extract the requested field
    parts = output.strip().split(":")
    INDEXES = {
        "username": 0,
        "uid": 2,
        "gid": 3,
        "gecos": 4,
        "home": 5,
        "shell": 6,
    }
    return parts[INDEXES[field]]


@typecheck
def add_alias(command: str) -> None:
    """
    Add an alias to ~/.bashrc if it doesn't already exist.

    :param command: The alias command to add
    :type command: str
    """
    alias_cmd = f"alias {command}"
    run_script_function(
        script=PACKAGE_BASH_HELPERS_PATH,
        function="write_marked",
        args=[alias_cmd, str(Path.home() / ".bashrc")],
        run=True,
        check=False,
    )
    log_step(f"Successfully added alias: {alias_cmd}")



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from argparse import ArgumentTypeError
from copy import deepcopy
from pathlib import Path

import networkx as nx

from gurk.lib.context import get_logger
from gurk.lib.context.registry import get_plugin_registration
from gurk.lib.shared.configs import load_toml, load_yaml
from gurk.lib.shared.dicts import pprint_typed_dict
from gurk.lib.shared.plugins import (
    PluginManifest,
    PluginMetadata,
    PluginMetadataProject,
    PluginOptions,
)
from gurk.lib.shared.remotes import is_git_repo
from gurk.lib.shared.scripts import (
    ScriptBlockTypes,
    check_script_blocks,
    get_block_spans,
)
from gurk.lib.shared.tasks import DefaultTaskDictCollection
from gurk.lib.utils import (
    GURK_MANIFEST_FILENAME,
    GURK_METADATA_FILENAME,
    PathLike,
    PatternCollection,
    check_version,
    full_isinstance,
    typecheck,
)

from .gurk_argparser import GurkArgumentParser, check_args_dict


@typecheck
def check_local_plugin(plugin_path: PathLike, verbose: bool = False) -> bool:
    """
    Check if a local plugin is valid.
        :NOTE: All imported plugins (recursively) must also be local and valid.

    :param plugin_path: Path to the local plugin directory
    :type plugin_path: PathLike
    :param verbose: Whether to print errors
    :type verbose: bool
    :return: True if the plugin is valid, False otherwise
    :rtype: bool
    """
    # Get logger
    logger = get_logger()

    # Directed graphs for checking cycles
    imports_graph = nx.DiGraph()
    task_dependency_graph = nx.DiGraph()
    task_supercedes_graph = nx.DiGraph()

    # Save available tasks
    available_tasks: DefaultTaskDictCollection = dict()

    def _check_local_plugin(_plugin_path: Path) -> bool:
        def error(message: str) -> bool:
            msg = f"'{_plugin_path}': {message}"
            if verbose:
                logger.error(msg)
            else:
                logger.debug(msg)

        # Check that the path exists and is a directory
        if not _plugin_path.exists():
            error("Plugin path does not exist.")
            return False
        elif not _plugin_path.is_dir():
            error("Plugin path is not a directory.")
            return False

        # Load pyproject.toml
        pyproject_file = _plugin_path / GURK_METADATA_FILENAME
        if not pyproject_file.is_file():
            error(
                f"Plugin source '{_plugin_path}' is missing a '{GURK_METADATA_FILENAME}' file."
            )
            return False
        pyproject_data = load_toml(pyproject_file)
        if not pyproject_data:
            error(
                f"Plugin source '{_plugin_path}' has an invalid '{GURK_METADATA_FILENAME}' file"
            )
            return False

        # Validate pyproject.toml
        if not full_isinstance(pyproject_data, PluginMetadata, extra="ignore"):
            error(
                f"Plugin source '{_plugin_path}' has an invalid '{GURK_METADATA_FILENAME}' "
                "file: invalid 'project' section structure. Expected (minimum):\n"
                f"{pprint_typed_dict(PluginMetadataProject, indent=2, as_str=True)}"
            )
            return False

        ## Validate version validity
        version = pyproject_data["project"]["version"]
        if not check_version(version):
            error(
                f"Plugin source '{_plugin_path}' has an invalid version "
                f"'{version}' in its '{GURK_METADATA_FILENAME}' file. Version "
                "should be in the format 'X.Y.Z' where X, Y, and Z are integers."
            )
            return False

        ## Valid and unique plugin name
        plugin_name = pyproject_data["project"]["name"]
        if not plugin_name.strip():
            error(
                f"Plugin source '{_plugin_path}' has an invalid "
                f"'{GURK_METADATA_FILENAME}' file: 'project.name' is empty."
            )
            return False
        elif not PatternCollection.NAMING.patterns.match(plugin_name):
            special_chars = ("-", "_")
            if plugin_name.startswith(special_chars) or plugin_name.endswith(
                special_chars
            ):
                error(
                    f"Plugin name '{plugin_name}' must "
                    "start and end with a lowercase letter."
                )
            else:
                error(
                    f"Plugin name '{plugin_name}' is invalid: It cannot have any "
                    "special characters except '_' or '-' and must be lowercase"
                )
            return False

        existing_plugin_registration = get_plugin_registration(
            plugin_name,
            home_registry=True,
            package_registry=True,
            require_local=False,
        )
        existing_plugin_entry = (
            next(iter(existing_plugin_registration.values()))
            if existing_plugin_registration
            else None
        )
        if existing_plugin_registration and (
            existing_plugin_entry.get("local") is not None
            and existing_plugin_entry["local"] != _plugin_path.resolve()
        ):
            existing_local = (
                existing_plugin_entry.get("local")
                if existing_plugin_entry
                else None
            )
            error(
                f"Plugin name '{plugin_name}' is already used by another plugin "
                f"at {existing_local if existing_local else '<no local path>'}."
            )
            return False

        # Load manifest file
        plugin: PluginManifest = load_yaml(
            _plugin_path / GURK_MANIFEST_FILENAME
        )
        if plugin is None:
            error(
                f"Plugin source '{_plugin_path}' has no '{GURK_MANIFEST_FILENAME}' file or it is invalid YAML."
            )
            return False

        # Validate structure
        plugin_without_helpers = {
            k: v
            for k, v in plugin.items()
            if isinstance(k, str) and not k.startswith("_")
        }
        if not full_isinstance(plugin_without_helpers, PluginManifest):
            error(
                f"Plugin at '{_plugin_path}' has invalid structure. Expected: "
                f"{pprint_typed_dict(PluginManifest, indent=2, as_str=True)}"
            )
            return False

        ## Check each task field
        plugin_tasks = plugin.get("tasks", {})
        for task_name, task in plugin_tasks.items():
            # Check task name
            plugin_prefix, remaining = (task_name.split("/", 1) + [None])[:2]
            if plugin_prefix != plugin_name or not remaining:
                error(
                    f"Task '{task_name}' has an invalid name. Its "
                    f"name should be '{plugin_name}/<task_name>'"
                )
                return False
            elif not PatternCollection.NAMING.patterns.match(remaining):
                special_chars = ("-", "_")
                if remaining.startswith(special_chars) or remaining.endswith(
                    special_chars
                ):
                    error(
                        f"Task '{task_name}' has an invalid subname: It "
                        "must start and end with a lowercase letter."
                    )
                else:
                    error(
                        f"Task '{task_name}' has an invalid subname: It cannot have any "
                        "special characters except '_' or '-' and must be lowercase"
                    )
                return False

            # Check task description
            if not task["description"]:
                error(f"Task '{task_name}' description is missing or empty.")
                return False

            # Check 'script' field
            ## Existence
            script = _plugin_path / task["script"]
            if not script.is_file():
                error(
                    f"Task '{task_name}' script file '{script}' does not exist."
                )
                return False
            ## Validity
            errors = check_script_blocks(script)
            if errors:
                error(
                    f"Task '{task_name}' script '{script}' has errors:\n"
                    + "\n".join(errors)
                )
                return False

            # Check 'function' field
            blocks = get_block_spans(script)
            if task["function"] is None:
                desired_block_type = ScriptBlockTypes.ENTRYPOINT
            else:
                desired_block_type = ScriptBlockTypes.FUNCTION
            if not any(
                b["name"] == task["function"]
                for b in blocks
                if b["type"] == desired_block_type
            ):
                f_or_e = (
                    ("function " + task["function"])
                    if task["function"]
                    else "entrypoint"
                )
                error(
                    f"Task '{task_name}' {f_or_e} does not exist in script '{script}'."
                )
                return False

            # Check 'config_file' field
            if task.get("config_file") is not None:
                config_file = _plugin_path / task["config_file"]
                if not config_file.is_file():
                    error(
                        f"Task '{task_name}' config file '{config_file}' does not exist."
                    )
                    return False

            # Check 'args' field
            task_args = task.get("args")
            if task_args:
                # Validate structure
                try:
                    check_args_dict(task_args)
                except ArgumentTypeError as e:
                    error(
                        f"Task '{task_name}' has invalid 'args' definition: {e}"
                    )
                    return False

                # Validate arg names
                arg_start = f"--{plugin_name}-"
                arg_names = [arg_name for arg_name in task_args.keys()]
                invalid_args = [
                    arg_name
                    for arg_name in arg_names
                    if not arg_name.startswith(arg_start)
                ]
                if invalid_args:
                    error(
                        f"Task '{task_name}' has an invalid arg names {invalid_args}. "
                        f"Arg names must be '{arg_start}<remaining>'."
                    )
                    return False

        # Check that the 'imports' section is valid
        imports = plugin.get("imports", [])
        if not full_isinstance(imports, list[str]):
            error(
                f"Plugin 'imports' section is not a list of strings, but of type '{type(imports)}'."
            )
            return False

        def _check_graph_cycles(graph: nx.DiGraph, field: str) -> bool:
            """
            Check for cycles in a directed graph.

            :param graph: The directed graph to check
            :type graph: nx.DiGraph
            :param field: The field being checked (for logging purposes)
            :type field: str
            :return: True if no cycles are found, False otherwise
            :rtype: bool
            """
            try:
                cycle = nx.find_cycle(graph)
                if cycle:
                    error(
                        f"Circular dependency detected in '{field}' field: {cycle}"
                    )
                    return False
            except nx.NetworkXNoCycle:
                # No cycle detected
                pass
            return True

        ## Check that imported plugins exist and are valid
        imports_graph.add_node(plugin_name)
        for imp in imports:
            # Check that imported plugins exist in the desired location
            if not get_plugin_registration(
                imp, home_registry=True, package_registry=True
            ):
                msg = f"Imported plugin '{imp}' does not exist locally."
                if is_git_repo(imp):
                    msg += f" You can pull it via 'gurk pull {imp}'."
                error(msg)
                return False

            # Check the imports graph for cycles
            imports_graph.add_edge(plugin_name, imp)
            if not _check_graph_cycles(imports_graph, "imports"):
                return False

            # Check imported plugin
            imp_registration = get_plugin_registration(
                imp, home_registry=True, package_registry=True
            )
            imp_local = next(iter(imp_registration.values()))["local"]
            if not _check_local_plugin(imp_local):
                error(f"Imported plugin '{imp}' is invalid.")
                return False

        # Add defined tasks to available tasks
        task_names = list(plugin_tasks.keys())
        available_tasks.update(plugin_tasks)

        def expand_graph(graph: nx.DiGraph, field: str) -> bool:
            """
            Expand a directed graph with edges from a task field.
                :NOTE: The direction is the same, as this is only used for cycle detection.

            :param graph: The directed graph to expand
            :type graph: nx.DiGraph
            :param field: The field from which to add edges
            :type field: str
            :return: True if successful, False if an unknown task is found
            :rtype: bool
            """
            graph.add_nodes_from(task_names)
            for task_name, task in plugin_tasks.items():
                for dep in task.get(field, []):
                    if dep not in available_tasks:
                        error(
                            f"Task '{task_name}' uses unknown "
                            f"task '{dep}' in '{field}' field."
                        )
                        return False
                    graph.add_edge(task_name, dep)

            return True

        # Check the task dependency graph
        dependency_field = "depends_on"
        if not expand_graph(task_dependency_graph, dependency_field):
            return False
        elif not _check_graph_cycles(task_dependency_graph, dependency_field):
            return False

        # Check the task supercedes graph
        supercedes_field = "supercedes"
        if not expand_graph(task_supercedes_graph, supercedes_field):
            return False
        elif not _check_graph_cycles(task_supercedes_graph, supercedes_field):
            return False

        # Check 'options' section
        options: PluginOptions = plugin["options"]
        if "default" not in options:
            error(
                "Plugin 'options' section is missing required 'default' option."
            )
            return False

        ## Validate each option
        for option_name, option in options.items():
            # Check option name
            if "/" in option_name or ":" in option_name:
                error(
                    f"Option name '{option_name}' is invalid. Option names cannot contain '/' or ':'."
                )
                return False

            # Check that all tasks in the option are defined
            for task_name in option.keys():
                if task_name not in available_tasks:
                    error(
                        f"Task '{task_name}' in '{option_name}' option "
                        "is not defined in this or any imported plugins."
                    )
                    return False

            # Check 'config_file' fields
            for task_name, task_spec in option.items():
                config_file = task_spec.get("config_file")
                if config_file:
                    config_file = _plugin_path / task_spec["config_file"]
                    if not config_file.is_file():
                        error(
                            f"Task '{task_name}' in option '{option_name}' "
                            f"config file '{config_file}' does not exist."
                        )
                        return False

            # Check that all args in the option are defined
            for task_name, task_spec in option.items():
                # Create a temporary parser to validate args
                parser = GurkArgumentParser(
                    add_verbose_arg=False,
                    add_non_interactive_arg=False,
                    add_force_arg=True,
                )
                allowed_args = available_tasks[task_name].get("args", {})
                parser.extend_arguments(allowed_args)

                # Override parser.error to raise exception instead of exiting
                def raise_error(message):
                    raise ValueError(message)

                parser.error = raise_error

                # Validate args
                try:
                    parser.parse_args(task_spec.get("args", []))
                except ValueError as e:
                    error(e)
                    return False

            # Determine enabled tasks (including dependencies)
            directly_enabled_tasks = set(option.keys())
            enabled_tasks = deepcopy(directly_enabled_tasks)
            for task_name in directly_enabled_tasks:
                enabled_tasks.update(
                    nx.descendants(task_dependency_graph, task_name)
                )

            # Check that at least one task is being run
            if not enabled_tasks:
                error(f"Option '{option_name}' has no enabled tasks.")
                return False

            # If any tasks are defined in the plugin, that at least one of them must be enabled
            if plugin_tasks and not set(plugin_tasks.keys()) & enabled_tasks:
                error(
                    f"Option '{option_name}' does not enable "
                    "any tasks defined in this plugin."
                )
                return False

            # Check that no two tasks that supercede each other are enabled together
            for u, v in task_supercedes_graph.edges():
                if u in enabled_tasks and v in enabled_tasks:
                    error(
                        f"Tasks '{u}' and '{v}' that supercede each other "
                        f"would both be enabled in the '{option_name}' option"
                    )
                    return False

        # All checks passed
        return True

    # Check plugin
    return _check_local_plugin(Path(plugin_path))



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import TextIOBase
from typing import Literal, overload

from rich.markup import escape

from gurk.lib.shared.printers import richprint
from gurk.lib.utils import typecheck

from .logger_types import LoggerSeverity


@typecheck
def _filter_pydantic_wrapper(traceback_str: str) -> str:
    """
    Filter out Pydantic's internal wrapper from error messages to improve readability.

    :param traceback_str: The original traceback string
    :type traceback_str: str
    :return: The filtered traceback string
    :rtype: str
    """
    lines = traceback_str.splitlines()
    cleaned = []

    i = 0
    while i < len(lines):
        # look ahead for a line containing _typecheck
        if i + 1 < len(lines) and "_typecheck(" in lines[i + 1]:
            # skip until we see validate_python
            while (
                i < len(lines)
                and "self.__pydantic_validator__.validate_python("
                not in lines[i]
            ):
                i += 1

            # skip the validate_python line and the line after it (if any)
            i += 2
            continue

        cleaned.append(lines[i])
        i += 1

    return "\n".join(cleaned)


@overload
def logrichprint(
    severity: LoggerSeverity,
    message: str,
    as_str: Literal[False] = ...,
    file: TextIOBase | None = ...,
) -> None: ...


@overload
def logrichprint(
    severity: LoggerSeverity,
    message: str,
    as_str: Literal[True] = ...,
    file: TextIOBase | None = ...,
) -> str: ...


@typecheck
def logrichprint(
    severity: LoggerSeverity,
    message: str,
    as_str: bool = False,
    file: TextIOBase | None = None,
) -> str | None:
    """
    Print a rich-formatted log message with the specified severity.

    :param severity: Severity level
    :type severity: LoggerSeverity
    :param message: The message to print
    :type message: str
    :param as_str: If True, return the formatted message as a string instead of printing it
    :type as_str: bool
    :param file: The output file (stdout/stderr). If None, defaults to stdout.
    :type file: TextIOBase | None
    :return: The formatted message if as_str is True, otherwise None
    :rtype: str | None
    """
    color = f"{'bold 'if severity.bold else ''}{'bright_'if severity.bright else ''}{severity.color}"
    msg = f"{richprint(escape(f'[{severity.label}]'), color=color, as_str=True)} {message}"
    if as_str:
        return msg
    else:
        richprint(msg, file=file)



# Copyright 2026 Arturo Roberti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
from importlib import resources
from pathlib import Path
from typing import Any

from ruamel.yaml.comments import CommentedMap, CommentedSeq

from gurk.lib.utils import PathLike, PatternCollection, typecheck

from .types import YAML, CommentedDict


@typecheck
def resolve_package_path(raw_script: PathLike, /) -> PathLike | None:
    """
    Resolve paths that may refer to package resources. Package paths are in the format:
    ```
    package://<package-name>/relative/path/inside/package
    ```
    The `package://` pattern does not need to be at the start of the string, and can appear
    multiple times in the string, but only the first occurrence of it is replaced.

    :param raw_script: Raw script path
    :type raw_script: PathLike
    :return: Resolved script path or None if package not found. The output type matches the input type.
    :rtype: PathLike | None
    """
    # Return wrong types as-is
    if not isinstance(raw_script, (Path, str)):
        return raw_script

    raw_str = str(raw_script)

    def _replace_package(match: re.Match) -> str:
        pkg_name, rel_path = match.groups()
        try:
            pkg_root = Path(resources.files(pkg_name))
        except ModuleNotFoundError:
            raise

        return str(pkg_root / rel_path)

    try:
        # Replace ALL occurrences of package://... anywhere in the string
        resolved_str = PatternCollection.PATH.patterns["package"].sub(
            _replace_package, raw_str
        )
    except ModuleNotFoundError:
        return None

    # NOTE: We use 'os' and no built-in 'Path' method to retain consecutive slashes
    resolved_str = os.path.expanduser(resolved_str)

    # Return same type as input
    if isinstance(raw_script, Path):
        return Path(resolved_str)
    else:
        return resolved_str


@typecheck
def load_yaml(
    yaml_file_or_str: PathLike,
    /,
    *,
    from_str: bool = False,
    remove_list_duplicates: bool = True,
) -> Any | None:
    """
    Load a YAML content and normalize it.

    :param yaml_file_or_str: Path to the YAML file to load or a string containing YAML content (if 'from_str' is True)
    :type yaml_file_or_str: PathLike
    :param from_str: Whether 'yaml_file_or_str' is a string containing YAML content (True) or a file path (False, default)
    :type from_str: bool
    :param remove_list_duplicates: Whether to remove duplicates in lists
    :type remove_list_duplicates: bool
    :return: Normalized content of the YAML file, or None if loading fails
    :rtype: Any | None
    """

    def normalize_yaml(obj: Any) -> Any:
        """
        Recursively normalize YAML content:
        - Convert all numbers to float
        - Remove duplicates in lists (if specified)
        - Resolve package paths in strings
        - Preserve comments and formatting using ruamel.yaml

        :param obj: The object to normalize
        :type obj: Any
        :return: Normalized object
        :rtype: Any
        """
        if not obj:
            # Empty or None
            return obj

        if isinstance(obj, CommentedMap):
            # Normalize values in-place to preserve comments
            for k, v in list(obj.items()):
                obj[k] = normalize_yaml(v)
            return obj

        elif isinstance(obj, CommentedSeq):
            if remove_list_duplicates:
                # Remove duplicates while preserving the original list
                seen = []
                indices_to_remove = []
                for i, item in enumerate(obj):
                    if item in seen:
                        indices_to_remove.append(i)
                    else:
                        seen.append(item)

                # Remove duplicates in reverse order to maintain indices
                for i in reversed(indices_to_remove):
                    del obj[i]

            # Normalize items in-place
            for i, item in enumerate(obj):
                obj[i] = normalize_yaml(item)
            return obj

        elif isinstance(obj, str):
            # Resolve package paths
            return resolve_package_path(obj)

        elif isinstance(obj, bool):
            # Keep booleans as-is
            return obj

        # ATTENTION: bool would evaluate as int here
        elif isinstance(obj, (int, float)):
            # Convert all numbers to float
            return float(obj)

        else:
            # Should not happen
            pass

    if from_str:
        try:
            content = YAML.load(yaml_file_or_str) or {}
        except Exception:
            return None
    else:
        if not Path(yaml_file_or_str).is_file():
            return None
        with open(yaml_file_or_str, "r") as f:
            try:
                content = YAML.load(f) or {}
            except Exception:
                return None
    return normalize_yaml(content)


@typecheck
def dump_yaml(content: CommentedDict[str, Any], yaml_file: PathLike) -> None:
    """
    Dump content to a YAML file.

    :param content: Content to dump
    :type content: CommentedDict[str, Any]
    :param yaml_file: Path to the YAML file to dump to
    :type yaml_file: PathLike
    """
    with open(yaml_file, "w", encoding="utf-8") as f:
        YAML.dump(content, f)
